
""" Generic Plotting tools """

import numpy as np
import matplotlib.pyplot as mpl

##############################
#                            #
#  MPL Tools                 #
#                            #
##############################
# ------------ #
#              #
#  Bar Plot    #
#              #
# ------------ #
def evolbar(timearray, data, ax=None, bottom=0,
            textbar=True, textloc="auto", fontsize=10, textincolor=None, textoutcolor=None,textformat=None,
            textprop={}, inpad=2, txtpad=0,  textmin=None, textadd="",
            yscale=None, formatxaxis=True,
            facecolors=None, edgecolors=None, alphas=None,
            clearaxis=True, clearwhich=["left","right","top"], **kwargs):
    """ 
    Parameters
    ----------

    // Text in Bar
    textbar: [bool] -optional-
        should the amplitude of the bar be shown in(out) the bar ?
        in or out depends on the size of the bar such that text fits in given its fontsize

    fontsize: [float] -optional-
        size of the font. string fonts are not allowed (i.e. large etc.)
        = ignored if textinbar is False =

    textincolor: [matplotlib's color] -optional-
        color of the text inside the bar.
        = ignored if textinbar is False =

    textloc: [string] -optional-
        where the text should be located ?
        - out: on top of the bar
        - in: inside the bar (on the top)
        - auto: 'in' or 'out' depending on the bar size.

    textprop: [dict] -optional-
        kwargs for ax.text() for the text inside the bar.
        
    inpad: [float] -optional-
        pad added to the top of the text inside the bar. In percent of ax size.
        = ignored if textbar is False =
        
    // Axes clearing
    clearaxis: [bool] -optional-
        should the axis lines and yticks be removed ?

    clearwhich [string or list of] -optional-
        which axis lines should be removed ? 
        - "*"/'all' -> ["bottom","left","right","top"]
        = ignored if clearaxis is not True =

    **kwargs goes to ax.bar()

    Returns
    -------
    fig, ax.bar's output
    """
    from matplotlib import dates as mdates
    #
    # - start: input check

    # - end: input check
    #
    if ax is None:
        fig = mpl.figure(figsize=[7,3])
        ax = fig.add_axes([0.1,0.2,0.8,0.7])
    else:
        fig = ax.figure

    if yscale is None:
        yscale = ax.get_ylim()[-1]
        
    if textbar:
        if textloc in ["auto"]:
            fontscale = ax.bbox.height/(2* yscale)
        elif textloc in ["in"]:
            fontscale = np.inf
        elif textloc in ["out"]:
            fontscale = 0
        else:
            raise ValueError(f'cannot parse the given textloc {textloc}: "auto", "in" or "out" allowed.')
    
    def _show_single_(ax_, time_, data_, bottom_, txtpad_=None, **kwargs):
        """ """
        artbar_ = ax_.bar(time_, data_, bottom=bottom_, **kwargs)
        
        # - Text
        if textbar:
            if len(np.atleast_1d(bottom_))==1:
                bottom_ = [np.atleast_1d(bottom_)[0]]*len(data_)

            prop     = dict(ha="center",zorder=8, fontsize=fontsize, weight="bold")
            prop_in  = {**prop, **dict(va="top"), **textprop}
            prop_out = {**prop, **dict(va="bottom"),**textprop}
            if txtpad_ is not None and len(np.atleast_1d(txtpad_))==1:
                txtpad_ = [np.atleast_1d(txtpad_)]*len(data_)
                
            txtlocation= []
            for i,(r_, t_,v_,b_) in enumerate(zip(artbar_.get_children(), time_, data_, bottom_)):
                text_ = v_ if textformat is None else f"{v_:{textformat}}"+textadd
                rect_fc = r_.get_facecolor()
                rect_ec = r_.get_edgecolor()
                has_fc = rect_fc[-1] > 0
                fc_w   = rect_fc[:3] == (1,1,1)
                has_ec = rect_ec[-1] > 0
                ec_w   = rect_ec[:3] == (1,1,1)
                
                if (textmin is not None and v_<textmin):
                    txtlocation.append("None")
                    continue

                lowerlimit = (fontsize+(txtpad_[i] if txtpad_ is not None else 0))
                if v_*fontscale>lowerlimit:
                    txtlocation.append("in")

                    if textincolor is None or textincolor in ["auto"]:
                        if has_fc and not fc_w:
                            textincolor_="w"
                        elif has_ec and not ec_w:
                            textincolor_=rect_ec
                        else:
                            textincolor_="w"
                    else:
                        textincolor_ = textincolor
                            
                    ax.text(t_, v_+b_  -yscale*inpad/100, text_, color=textincolor_, **prop_in)
                else:           
                    txtlocation.append("out")
                    ax.text(t_,  v_+b_, text_, color=rect_fc if textoutcolor is None else textoutcolor , **prop_out)
        else:
            txtlocation = None
            
        return artbar_, txtlocation
              
    #    
    # = Start: Plotting Data
    
    # Single Data
    if len(np.shape(data))==1:
        # - Bar
        artbar, txtlocation = _show_single_(ax, time_=timearray, data_=data, bottom_=bottom)
        
    # List of Data
    else:
        cdata = np.cumsum(data, axis=0)
        yscale = np.max(cdata)*1.1
        fontscale = ax.bbox.height/(2* yscale)
        
        artbar = []
        for i,data_ in enumerate(data):
            prop = {**_read_prop_("alpha",alphas, i),
                    **_read_prop_("facecolor",facecolors, i),
                    **_read_prop_("edgecolor",edgecolors, i),
                    **kwargs}
            if i == 0:
                txtpad_ = None
                bottom_ = 0
            else:
                txtpad_ = np.ones(len(data_))*fontsize
                txtpad_[~np.in1d(txtlocation,["out"])] = 0
                bottom_ = cdata[i-1]                    
            artbar_, txtlocation = _show_single_( ax, time_=timearray, 
                                                        data_=data_, bottom_=bottom_,
                                                        txtpad_=txtpad_,**prop)
            artbar.append(artbar_)
            
    # = end: Plotting Data
    #
    
    ax.set_ylim(0, yscale)
    #
    # - start: time formating
    if formatxaxis:
        locator = mdates.AutoDateLocator()
        formatter = mdates.ConciseDateFormatter(locator)
        ax.xaxis.set_major_locator(locator)
        ax.xaxis.set_major_formatter(formatter)
    # - end: time formating
    #

    if clearaxis and clearwhich is not None:
        ax.get_yaxis().set_visible(False)
        if type(clearwhich) is str:
            if clearwhich in ["all","*"]:
                clearwhich = ["bottom","left","right","top"]
            else:
                clearwhich = np.atleast_1d(clearwhich)
                
        [ax.spines[which].set_visible(False) for which in clearwhich]

    return fig, artbar


def _read_prop_(k,v, ith=None):
    """ """
    if v is None:
        return {}
    if len(np.atleast_1d(v))==1:
        return {k:np.atleast_1d(v)[0]}
    return {k:np.atleast_1d(v)[ith]}
# ------------ #
#              #
#  ColorBar    #
#              #
# ------------ #

def colorbar(ax, cmap, vmin=0, vmax=1, label="",
             fontsize="x-large", **kwargs):
    """ Set a colorbar in the given axis

    Parameters
    -----------
    ax: [mpl's Axes]
        Axis in which the colorbar will be drawn

    cmap: [mpl's colormap]
        A matplotlib colormap

    vmin, vmax: [float,float] -optional-
        Extend of the colormap, values of the upper and lower colors

    label, fontsize: [string, string/float] -optional-
        Label of the colorbar and its associated size
     
    **kwargs goes to matplotlib.colobar.ColorbarBase

    Return
    ------
    colorbar
    """
    import matplotlib
    
    if "orientation" not in kwargs.keys():
        bbox = ax.get_position()
        orientiation = "vertical" if bbox.xmax - bbox.xmin < bbox.ymax - bbox.ymin \
          else "horizontal"
        kwargs["orientation"] = orientiation

    norm    = matplotlib.colors.Normalize(vmin=vmin,vmax=vmax)
    c_bar   = matplotlib.colorbar.ColorbarBase(ax, cmap=cmap,
                              norm=norm,**kwargs)
    
    c_bar.set_label(label,fontsize=fontsize)
    if "ticks" in kwargs.keys() and "ticklabels" not in kwargs.keys():
        c_bar.ax.set_xticklabels([r"%s"%v for v in kwargs["ticks"]])
        
    return c_bar

def insert_ax(ax, location, shrunk=0.7, space=.05,
              axspace=0.02, shareax=False, **kwargs):
    """ insert an axis at the requested location

              
    The new axis will share the main axis x-axis (location=top or bottom) or
    the y-axis (location=left or right).

    Parameters:
    -----------
    location: [string]
       top/bottom/left/right, i.e. where new axis will be set

    shrunk: [float]
        the main axis will be reduced by so much (0.7 = 70%).
        the new axis will take the room

    space: [float]
        extra space new axis does not use between it and the edge of
        the figure. (in figure unit, i.e., [0,1])

    axspace: [float]
        extra space new axis does not use between it and the input
        axis. (in figure unit, i.e., [0,1])

    shareax: [bool]
        The new axis will share the main axis x-axis (location=top or bottom) or
        the y-axis (location=left or right). If so, the axis ticks will be cleaned.
                           
    **kwargs goes to figure.add_axes() for the new axis

    Returns:
    --------
    axes (the new axis)
    """
    from matplotlib.transforms  import Bbox
    # --------------------
    # hist x
    # -------------------- #
    # -- keep trace of the original axes
    bboxorig = ax.get_position().frozen()

    if location in ["top","bottom"]:
        axhist = ax.figure.add_axes(np.random.rand(4),
                                    sharex=ax if shareax else None,
                                    **kwargs) # This will be changed
        _bboxax = ax.get_position().shrunk(1,shrunk)
        _bboxhist = Bbox([[_bboxax.xmin, _bboxax.ymax+axspace ],
                          [_bboxax.xmax, bboxorig.ymax-space]])
        
        if location == "bottom":
            tanslate = _bboxhist.height + space+axspace
            _bboxhist = _bboxhist.translated(0, bboxorig.ymin-_bboxhist.ymin+space)
            _bboxax = _bboxax.translated(0,tanslate)
            
    # --------------------
    # hist y
    # -------------------- #            
    elif location in ["right","left"]:
        axhist = ax.figure.add_axes([0.5,0.1,0.2,0.42],sharey=ax if shareax else None,
                                    **kwargs) # This will be changed
        _bboxax = ax.get_position().shrunk(shrunk,1)
        _bboxhist = Bbox([[_bboxax.xmax+axspace, _bboxax.ymin ],
                          [bboxorig.xmax-space, _bboxax.ymax]])
        if location == "left":
            tanslate = _bboxhist.width + space + axspace
            _bboxhist = _bboxhist.translated(bboxorig.xmin-_bboxhist.xmin+space, 0)
            _bboxax = _bboxax.translated(tanslate,0)
        
    else:
        raise ValueError("location must be 'top'/'bottom'/'left' or 'right'")


    axhist.set_position(_bboxhist)
    ax.set_position(_bboxax)

    # ---------------------
    # remove their ticks
    if shareax:
        if location in ["top","right"]:
            [[label.set_visible(False) for label in lticks]
            for lticks in [axhist.get_xticklabels(),axhist.get_yticklabels()]]
        elif location == "bottom":
            [[label.set_visible(False) for label in lticks]
            for lticks in [ax.get_xticklabels(),axhist.get_yticklabels()]]
        elif location == "left":
            [[label.set_visible(False) for label in lticks]
            for lticks in [ax.get_yticklabels(),axhist.get_xticklabels()]]
    
    return axhist


# ---------------- #
#                  #
#   HistColorbar   #
#                  #
# ---------------- #
class HistColorbar():
    def __init__(self, data=None, ax=None,  cax=None, fig=None, cmap=None, 
                 bins=10, vmin=None, vmax=None, sequence=None, histprop={},
                 draw=True, **kwargs):
        """ """
        self.out = {}
        self.set_ax(ax=ax, cax=cax, fig=fig)
        if data is not None:
            self.build_histrogram(data=data, bins=bins, vmin=vmin, vmax=vmax, **histprop)
        
        self.load_cmap(cmap, sequence=sequence)
            
        if draw:
            self.draw(**kwargs)
            
    # ============= #
    #   Methods      #
    # ============= #
    def set_ax(self, ax=None, cax=None, fig=None):
        """ """
        # ax=None, cax=None
        if ax is None and cax is None:
            if fig is None:
                self._fig = mpl.figure(figsize=[8,1])
            self._ax  = self.fig.add_axes([0.1,0.435,0.8,0.5])
            self._cax = self.fig.add_axes([0.1,0.2,0.8,0.1])
            
        # ax=None, cax is not
        elif ax is None:
            if len(np.atleast_1d(cax))==4:
                if fig is None:
                    self._fig = mpl.figure(figsize=[8,1])
                self._cax = self.fig.add_axes(cax)
            else:
                self._cax = cax
                self._fig = self.cax.figure

            self._ax = None
            
        # ax is not, cax=None
        elif cax is None:
            if len(np.atleast_1d(ax))==4:
                if fig is None:
                    self._fig = mpl.figure(figsize=[8,1])
                self._ax = self.fig.add_axes(ax)
            else:
                self._ax  = ax
                self._fig = self.ax.figure
                
            self._fig = self._ax.figure # in case
            self._cax = None

        else:
            if len(np.atleast_1d(ax))==4:
                if fig is None:
                    fig  = mpl.figure(figsize=[8,1])
                self._ax = fig.add_axes(ax)
            else:
                self._ax = ax
                
                
            if len(np.atleast_1d(cax))==4:
                if fig is None:
                    fig   = mpl.figure(figsize=[8,1])
                self._cax = fig.add_axes(cax)
            else:
                self._cax = cax
                
            self._fig = self.ax.figure

    def set_data(self, data, set_vrange=True):
        """ """
        data = np.asarray(data)
        self._data = data[data==data]

    def set_vrange(self, vmin, vmax):
        """ """
        if vmin is None and self.has_data():
            vmin = np.percentile(self.data, 2)
        if vmax is None and self.has_data():
            vmax = np.percentile(self.data, 98)

        self._vrange = [vmin, vmax]

    def set_visible(self, bool_):
        """ """
        if self._ax is not None:
            self.ax.set_visible(bool_)
            
        if self._cax is not None:
            self.cax.set_visible(bool_)
            
    def load_cmap(self, cmap=None, sequence=None):
        """ """
        if sequence == "None":
            sequence = None # No sequence
        elif sequence is None and self.has_bins():
            sequence = self.bins["size"]

        self._cmap = mpl.cm.get_cmap(cmap, sequence)
        
    def build_histrogram(self, data=None, vmin=None, vmax=None, bins=None, **kwargs):
        """ """
        if data is not None:
            self.set_data(data)
            
        self.set_vrange(vmin, vmax)
        if bins is None:
            bins = "auto"

        self._intensity, binegdes = np.histogram(self.data, range=self.vrange, bins=bins,  **kwargs)
        self._bins = {"edge":binegdes,
                      "centroid":np.mean([binegdes[1:],binegdes[:-1]], axis=0),
                      "width":binegdes[1:]-binegdes[:-1],
                      "size":len(binegdes)-1,
                      "vmin":binegdes[0],
                      "vmax":binegdes[-1],
                     }
        
        
    def draw(self, ax=None, xscale=True, swicth_offaxis=True, cbarprop={}, alpha=0.8, **kwargs):
        """ """
        if ax is not None:
            self.set_ax(ax)

        if self.has_unique_value():
            if self.ax is not None:
                self.ax.text(0.5,0.5, f"all to {np.unique(self.data)[0]}", va="center", ha="center",
                             transform=self.ax.transAxes, fontsize="x-small", color="0.7")
                self.ax.axis("off")
            if self.cax is not None:
                self.cax.set_visible(False)
            return
        
        # - Show Data
        elif self.has_hist() and self.ax is not None :
            if "colors" not in self.bins:
                self.bins["colors"] = self.cmap((self.bins["centroid"] - self.bins["vmin"])/(
                                             self.bins["vmax"]     - self.bins["vmin"])
                                            )
            self.out["bar"] = self.ax.bar(self.bins["centroid"], self.intensity, 
                           width=self.bins["width"], color=self.bins["colors"],
                           alpha=alpha,
                           **kwargs)
        
            self.ax.set_ylim(bottom=0)
            
            if xscale:
                self.ax.set_xlim(*self.vrange)
                if swicth_offaxis:
                    self.ax.axis("off")
                
        # - Show colorbar                
        if self.cax is not None:
            if self.ax is not None:
                self.ax.set_xticks([])

            if self.has_bins():
                cbarprop["vmin"]=self.vrange[0]
                cbarprop["vmax"]=self.vrange[1]

            self.out["cbar"] = colorbar(self.cax, self.cmap, **cbarprop)
        

    def set_label(self, label, fontsize=None, **kwargs):
        """ """
        if "cbar" in self.out:
            self.out["cbar"].set_label(label,fontsize=fontsize, **kwargs)
        else:
            self.ax.set_xlabel(label,fontsize=fontsize, **kwargs)
            
    def set_xticks(self, where="centroid"):
        """ """
        if type(where) is str:
            if where in ["center","centroid","centers","centroids"]:
                return self.ax.set_xticks(self.bins["centroid"])
            if where in ["edge","edges"]:
                return ax.set_xticks(self.bins["edge"])
                
            raise ValueError(f"Cannor parse 'where' to set the ticks {where} given")
        else:
            self.ax.set_xticks(where)
                            
    # ============= #
    #  Properties   #
    # ============= #
    @property
    def ax(self):
        """ """
        return self._ax

    @property
    def cax(self):
        """ """
        return self._cax

    @property
    def fig(self):
        """ """
        return self._fig
    
    @property
    def data(self):
        """ """
        return self._data if hasattr(self,"_data") else None

    def has_data(self):
        """ """
        return hasattr(self,"_data") and self._data is not None

    @property
    def vrange(self):
        """ """
        return self._vrange if hasattr(self,"_vrange") else None
    
    @property
    def intensity(self):
        """ """
        return self._intensity if hasattr(self,"_intensity") else None

    def has_hist(self):
        """ """
        return hasattr(self,"_intensity") and self._intensity is not None

    @property
    def bins(self):
        """ dict """
        return self._bins if hasattr(self,"_bins") else {}

    def has_bins(self):
        """ """
        return hasattr(self,"_bins") and self._bins is not None and len(self._bins)>0

    def has_unique_value(self):
        """ """
        return self.has_hist() and len(np.unique(self._intensity))==1
    
    @property
    def cmap(self):
        """ """
        return self._cmap if hasattr(self,"_cmap") else None



#!/usr/bin/env python
#

""" Library containing field information """
import os, sys
import pandas
import numpy as np
import warnings
from pandas import read_csv
from astropy import units, coordinates, time
import matplotlib.pyplot as mpl
from matplotlib.patches import Polygon




_FIELD_SOURCE = os.path.dirname(os.path.realpath(__file__))+"/data/ztf_fields.txt"
FIELD_DATAFRAME = read_csv(_FIELD_SOURCE, index_col="ID")
FIELDSNAMES = FIELD_DATAFRAME.index.values

_CCD_COORDS  = read_csv(os.path.dirname(os.path.realpath(__file__))+"/data/ztf_ccd_layout.tbl").rename(columns={"CCD ":"CCD"}) # corner of each CCDS

FIELD_COLOR = {1: "C2", 2: "C3", 3:"C1"}
FIELD_CMAP = {1: mpl.cm.Greens, 2:mpl.cm.Reds, 3:mpl.cm.Oranges}
FIELDNAME_COLOR = {"zg": "C2", "zr":"C3", "zi":"C1"}

_PLOTORIGIN = 180

def _load_fields_geoserie_(inclccd=False):
    """ Loads the FIELDS_GEOSERIE global variable 
    = Internal Tools =
    """
    if not inclccd:
        global FIELDS_GEOSERIE
    else:
        global FIELD_CCDS_GEOSERIE
    
    try:
        from geopandas import geoseries
    except ImportError:
        warnings.warn("You do not have geopandas, Please run pip install geopandas.")
        if not inclccd:
            FIELDS_GEOSERIE = None
        else:
            FIELD_CCDS_GEOSERIE = None 
        return
    
    field_verts = get_field_vertices(fieldid=FIELDSNAMES, inclccd=inclccd, asdict=True, aspolygon=True)
    if not inclccd:
        FIELDS_GEOSERIE = geoseries.GeoSeries(field_verts)
    else:
        FIELD_CCDS_GEOSERIE = geoseries.GeoSeries(field_verts)

    
def get_fields_geoserie(inclccd=False):
    """ returns the global variable FIELDS_GEOSERIE, creates it if necessary """
    if not inclccd:
        if not hasattr(sys.modules[__name__], 'FIELDS_GEOSERIE'):
            _load_fields_geoserie_(inclccd=False)
        
        return FIELDS_GEOSERIE
    else:
        if not hasattr(sys.modules[__name__], 'FIELD_CCDS_GEOSERIE'):
            _load_fields_geoserie_(inclccd=True)
        
        return FIELD_CCDS_GEOSERIE

# ------------------ #
#                    #
# Generic Tools      #
#                    #
# ------------------ #
def get_fieldid(grid=None, decrange=None, rarange=None, 
                gallrange=None, galbrange=None, 
                ecllrange=None, eclbrange=None, 
                ebvrange=None, verbose=False):
                
    """ 

    Parameters
    ----------
    grid: [None/str] -optional-
        Select the grid you want. If None the both will be considered.
        'main' or 'secondary' expected otherwise.

    decrange, rarange, galbrange, gallrange, ecllgrange, eclbrange: [2-array or None] -optional-
        dec, ra, galactic l, galactic b, ecliptic long and ecliptic lat to be considered [inclusive]. 
        In degree.
        3 formats available:
        - None: means no selection
        - [min,max]: means range to be considered, None means no limit. 
          example: decrange=[-10,None] means dec>-10.
        - [[min1,max1],[min2,max2], ...]: means zones to be considered.
          example: decrange=[[None, -10],[5,None]] will simply exclude the [-10,5] dec range.
        - None means no selection.

    ebvrange: [2-array or None] -optional-
        same format as 'decrange' but for Mily way E(B-V) extinction.
        
    """
    def get_query(key, krange=None, merging_logic="or"):
        """ """
        def _build_2d_(kmin,kmax):
            """ """
            if kmin is None and kmax is None:
                return []
            if kmin is None:
                return [f"{key}<={kmax}"]
            elif kmax is None:
                return [f"{key}>={kmin}"]
            else:
                return [f"{kmin}<={key}<={kmax}"]
        
        if krange is None:
            return []
        if np.shape(krange) == (2,):
            return _build_2d_(*krange)
        if np.shape(krange) == (2,2):
            query = [_build_2d_(*krange_) for krange_ in krange]
            return ["("+f" {merging_logic} ".join(np.squeeze(query))+")"]
        
        raise ValueError(f"Cannot for the format of the input krange: {krange}")

    query = []
    # Grid
    gridid = None if (grid is None or grid in ["*","all"]) else get_grid_field(grid)
    query.append([] if gridid is None else ["ID in @gridid"])
    # Ra
    
    query.append(get_query("RA", rarange))
    # Dec
    query.append(get_query("Dec", decrange))
    # gall
    query.append(get_query("GalLong", gallrange))
    # galb
    query.append(get_query("GalLat", galbrange))
    # ecll
    query.append(get_query("EclLong", ecllrange))
    # eclb
    query.append(get_query("EclLat", eclbrange))
    # MW ebmv
    query.append(get_query("Ebv", ebvrange))
    
    
    queries = np.concatenate(query)
    if verbose:
        print(queries)
    if len(queries) == 0:
        return FIELD_DATAFRAME.index
    if verbose:
        print(" & ".join(queries) ) 
    return FIELD_DATAFRAME.query( " & ".join(queries) ).index


def get_field_ccd_qid(ra, dec):
    """ """
    d_ = {}
    field_ccds = get_fields_containing_target(ra,dec, inclccd=True)
    for field_ccd in field_ccds:
        (ramin,decmin),(ramax,decmax) = np.percentile(np.asarray(FIELD_CCDS_GEOSERIE[field_ccd].exterior.xy).T, [0,100], axis=0)
        ccdpos=np.asarray([(ra-ramin)/(ramax-ramin)*6144,(dec-decmin)/(decmax-decmin)*6160])
        qid = ccdpos_to_qid(*ccdpos)
        field, ccd = field_ccd.split("_")
        d_[int(field)] = {"ccd":int(ccd), "qid":int(qid), "rcid":ccdid_qid_to_rcid(int(ccd), int(qid))}
        
    return d_
    
def get_fields_containing_target(ra, dec, inclccd=False):
    """ return the list of fields into which the position ra, dec is. 
    Remark that this is based on predefined field positions. 
    Hence, small attrition could affect this.

    Parameters
    ----------
    ra,dec: [float,float]
       coordinates in degree. 
    
    inclccd: [bool] -optional-
        do you want the details of the CCD id on top of the field
        format: "field_ccdid"

    Returns
    -------
    list (all the field ID that contain the given ra,dec coordinates)
    """
    try:
        from shapely import geometry
    except ImportError:
        raise ImportError("You need shapely to use this function. pip install shapely")
    
    
    coordpoint = geometry.Point(ra, dec)
    fields_geoserie = get_fields_geoserie(inclccd=inclccd)
    if fields_geoserie is None:
        warnings.warn("get_fields_containing_target would be much faster if you install geopandas (pip install geopandas)")
        return [f for f in FIELDSNAMES
                if geometry.Polygon( get_field_vertices(f)[0]).contains(coordpoint)]
    
    return fields_geoserie.index[ fields_geoserie.contains(coordpoint) ]


def get_field_vertices(fieldid=None, inclccd=False, ccd=None, asdict=False, aspolygon=False,
                        squeeze=True):
    """ Get the fields countours 
    
    Parameters
    ----------
    fieldid: [string or list of] -optional-
        Field (or list of) names as int. 
        If None, all the fields will be used.

    // output format

    asdict: [bool] -optional-
        Do you want to result as a list (asdict=False) following the input's fieldid sorting
        or do you want a dict ({fieldid_: verts_ ... })

    aspolygon: [bool] -optional-
        Do you want the vertices as 2d-array (aspolygon=False) or as shapely Geometries (True)

    squeeze: [bool] -optional-
        Should unnecessary dimension be removed ?
        if asdict is True:
            if squeeze: {fid_ccdid: }
            if not squeeze: {fid:{ccdid: }}
            = if not inclccd, squeez ignored =
        if not asdict:
           using np.squeeze, basically doing [[]]->[]
    Returns
    -------
    list of dict (see asdict)
    """
    if fieldid is None:
        fieldid = FIELDSNAMES
        
    if inclccd and ccd is None:
        ccd = np.arange(1,17)

    # - Actual calculation
    rafields, decfields  = get_field_centroid( np.asarray(np.atleast_1d(fieldid), dtype="int") ).T
    fields_verts = get_corners(rafields, decfields, inclccd=inclccd, ccd=ccd,
                                       inrad=False, squeeze=False)

    # ----------- #
    #  Format     #
    # ----------- #    
    if aspolygon:
        try:
            from shapely import geometry
            fields_countours = [[geometry.Polygon(fields_verts[f_][c_])
                                         for c_,ccd_ in enumerate(np.atleast_1d(ccd))]
                                         for f_,field_ in enumerate(np.atleast_1d(fieldid))]
        except ImportError:
            warnings.warn("You do not have shapely, Please run pip install shapely. 'aspolygon' set to False")
    else:
        fields_countours = fields_verts

    # ----------- #
    #  Output     #
    # ----------- #    
    if not asdict:
        return fields_countours if not squeeze else np.squeeze(fields_countours)

    # full camera dict
    if not inclccd:
        return {i:k[0] for i,k in zip(fieldid,fields_countours)}
    
    # ccd dict
    if squeeze:
        return {f"{field_}_{ccd_}":fields_countours[f_][c_]
                    for c_,ccd_ in enumerate(np.atleast_1d(ccd)) for f_,field_ in enumerate(np.atleast_1d(fieldid))}
    else:
        return {field_:{ccd_:fields_countours[f_][c_] for c_,ccd_ in enumerate(np.atleast_1d(ccd))}
                    for f_,field_ in enumerate(np.atleast_1d(fieldid))}

def get_field_centroid(fieldid, system="radec"):
    """ Returns the central coordinate [RA,Dec] or  of the given field 

    Parameters
    ----------
    fieldid: [int]
        single field ID

    system: [string] -optional-
        which coordinate system ?
        radec / galactic / ecliptic (default radec)

    Returns
    -------
    [[x_i, y_i],[]]... (depending on your coordinate system)
    Remark if only 1 fieldid given, you have [[x,y]] (not [x,y])
    """
    if system in ["radec", "RADec","RA,Dec", "ra,dec"]:
        syst = ["RA", "Dec"]
    elif system.lower() in ["gal","galactic"]:
        syst = ["Gal Long","Gal Lat"]
    elif system.lower() in ["ecl","ecliptic"]:
        syst = ["Ecl Long","Ecl Lat"]
    else:
        raise ValueError("unknown coordinate system %s select among: [radec / galactic / ecliptic]"%system)
    fieldid = np.atleast_1d(fieldid)
    radec = np.asarray(FIELD_DATAFRAME[np.in1d(FIELD_DATAFRAME.index, fieldid)][syst].values)
    
    return radec

def get_corners(ra_field, dec_field, inclccd=False, ccd=None, steps=5, squeeze=True, inrad=False):
    """ """
    from .utils.tools import rot_xz_sph, _DEG2RA
    
    if not inclccd:
        upper_left_corner = _CCD_COORDS.max()
        lower_right_corner = _CCD_COORDS.min()
    elif ccd is None:
        upper_left_corner = _CCD_COORDS.groupby("CCD").max()
        lower_right_corner = _CCD_COORDS.groupby("CCD").min()
    else:
        upper_left_corner = _CCD_COORDS.groupby("CCD").max().loc[ccd]
        lower_right_corner = _CCD_COORDS.groupby("CCD").min().loc[ccd]
        
    ewmin = -np.atleast_1d(upper_left_corner["EW"])
    nsmax = np.atleast_1d(upper_left_corner["NS"])
    ewmax = -np.atleast_1d(lower_right_corner["EW"])
    nsmin = np.atleast_1d(lower_right_corner["NS"])

    ra1  = (np.linspace(ewmax, ewmin, steps)/np.cos(nsmax*_DEG2RA)).T
    dec1 = (np.ones((steps,1))*nsmax).T
    #
    dec2  = np.linspace(nsmax,nsmin, steps).T
    ra2   = ewmin[:,None]/np.cos(dec2*_DEG2RA)
    #
    ra3 = (np.linspace(ewmin,ewmax, steps)/np.cos(nsmin*_DEG2RA)).T
    dec3 = (np.ones((steps,1))*nsmin).T
    #
    dec4  = np.linspace(nsmin,nsmax, steps).T
    ra4 = ewmax[:,None]/np.cos(dec4*_DEG2RA)

    ra_bd = np.concatenate((ra1, ra2, ra3, ra4  ), axis=1)  
    dec_bd = np.concatenate((dec1, dec2, dec3,dec4 ), axis=1)
    
    ra,dec = rot_xz_sph(np.moveaxis(ra_bd,0,1), np.moveaxis(dec_bd,0,1), np.moveaxis(np.atleast_3d(dec_field),0,1))
    ra += np.moveaxis(np.atleast_3d(ra_field),0,1)

    if inrad:
        ra *= _DEG2RA
        dec *= _DEG2RA
        
    radec = np.moveaxis([ra,dec],(0,1,2,3),(3,0,2,1))
    return radec if not squeeze else np.squeeze(radec)

def get_grid_field(which):
    """ """
    if which in ["main","Main","primary"]:
        return FIELDSNAMES[FIELDSNAMES<880]
    if which in ["aux","secondary", "auxiliary"]:
       return FIELDSNAMES[FIELDSNAMES>999]
    if which in ["all","*","both"]:
        return FIELDSNAMES
        
    raise ValueError(f"Cannot parse which field grid you want {which}")


def get_rcid_centroid(rcid, fieldid):
    """ """
    ccdid, q1d = rcid_to_ccdid_qid(rcid)
    return get_qids_centroid(fieldid, ccdid)[f"q{q1d}"]

def get_qids_centroid(fieldid=None, ccdid=None, ccd_vertices=None):
    if ccd_vertices is None:
        if fieldid is None or ccdid is None:
            raise ValueError("either fieldid and ccdid or ccd_vertices should be given")
        
        ccd_vertices = get_field_vertices(fieldid=fieldid, inclccd=True, ccd=ccdid)
        
    min_,mean_,max_ = np.percentile(ccd_vertices, [0,50,100], axis=0)
    return {"q1":np.mean([mean_, max_], axis=0),
            "q2":[np.mean([mean_[0], min_[0]]),np.mean([mean_[1], max_[1]])],
            "q3":np.mean([mean_, min_], axis=0),
            "q4":[np.mean([mean_[0], max_[0]]),np.mean([mean_[1], min_[1]])]
            }
                



def ccdpos_to_qid(ccdx, ccdy):
    """ returns the qid for the given ccd position """
    flagqid = np.asarray(np.asarray([ccdx<3072, ccdy<3080]), dtype=int)    
    return np.asarray([[1,4],[2,3]])[flagqid[0]][flagqid[1]]

def ccdid_qid_to_rcid(ccdid, qid):
    """ computes the rcid """
    return 4*(ccdid - 1) + qid - 1

def rcid_to_ccdid_qid(rcid):
    """ computes the rcid """
    qid = (rcid%4)+1
    ccdid  = int((rcid-(qid - 1))/4 +1)
    return ccdid,qid

##############################
#                            #
#  Fields and References     #
#                            #
##############################
def has_field_reference(fieldid, rcid_details=False, **kwargs):
    """ get the following dictionary {zg:bool, zr:bool, zi:bool}
    where bool is True if the field has a reference image and false otherwise
    
    **kwargs goes to load_metadata(), for instance auth=[username, password]
    Returns
    -------
    {zg:bool, zr:bool, zi:bool}
    """
    from .query import ZTFQuery
    zquery_ = ZTFQuery()
    zquery_.load_metadata(kind="ref", sql_query=f"field={fieldid}", **kwargs)
    if rcid_details:
        return {k:zquery_.metatable.query(f"filtercode in ['{k}']")["rcid"].value_counts(sort=False).to_dict() for k in ["zg", "zr","zi"]}
    return {k: k in zquery_.metatable["filtercode"].values for k in ["zg", "zr","zi"]}

def get_fields_with_band_reference(filter_, ccdid=1, qid=1, **kwargs):
    """ returns the list of fieldid that have a reference image in the `filter_` band.
    filter_ is a filtercode entry [zg, zr or zg]
    
    **kwargs goes to load_metadata(), for instance auth=[username, password]
    Returns
    -------
    list of fieldid
    """
    from .query import ZTFQuery
    zquery_ = ZTFQuery()
    zquery_.load_metadata(kind="ref",
            sql_query="filtercode='%s' and ccdid=%s and qid=%s"%(filter_,ccdid,qid), **kwargs)
    return zquery_.metatable["field"].values

def show_reference_map(band, **kwargs):
    """ Display the 'field plot' in which field with image reference in the given band are colored. """
    title   = "Fields with reference in the %s-band"%band[1]
    field_i = get_fields_with_band_reference(band)
    return show_fields(field_i, facecolor=FIELDNAME_COLOR[band], alpha=0.3, title=title, **kwargs)
    
# ===================== #
#                       #
#    SHOW FIELD         #
#                       #
# ===================== #
def show_fields(fields, vmin=None, vmax=None,
                ax=None, cmap="viridis", title=None,
                colorbar=True, cax=None, hcax=None, clabel=" ", inclhist=True, 
                show_ztf_fields=True, grid="main", grid_prop={},
                bkgd_fields=None, bkgd_prop={},
                show_mw=True, mw_b=None, mw_prop={},
                savefile=None, figsize=None,
                axparam={}, get_fplot=False,
                **kwargs):
    """ 
    Parameters
    ----------
    colored_by: 
    """
    fplot = FieldPlotter(ax=ax, cax=cax, hcax=hcax, figsize=figsize, inclcax=colorbar,
                             inclhist=inclhist, **axparam)
    # - Plotting
    if show_ztf_fields:
        fplot.show_ztf_grid(which=grid, **grid_prop)
        
    if bkgd_fields is not None:
        def_prop = dict(facecolor="0.7", edgecolor="0.5", alpha=0.2, zorder=2)
        fplot.show_fields(bkgd_fields,**{**def_prop,**bkgd_prop})
        
    if show_mw:
        fplot.show_milkyway(b=mw_b, **mw_prop)

    # Removing the NaNs
    fplot.show_fields(fields,
                        colorbar=colorbar,
                        clabel=clabel,cmap=cmap,
                        vmin=vmin, vmax=vmax,**kwargs)
    
    if title is not None:
        fplot.fig.text(0.5,0.9, title,
                     va="top", ha="center", fontsize="large")
    # Output
    if savefile is not None:
        fplot.fig.savefig(savefile, dpi=150)
    if get_fplot:
        return fplot
    return fplot.fig

def show_field_ccds(fieldid, ax=None, ccd=None, textcolor="k", facecolor="0.9", edgecolor="k",
                        autoscale=True, **kwargs):
    """ """
    if ax is None:
        fig = mpl.figure(figsize=[8,6])
        ax = fig.add_subplot(111)
    else:
        fig = ax.figure

    if ccd is None:
        ccd = range(1,17)
    ff = get_fields_geoserie(inclccd=True)
    fccd = {i:ff[f"{fieldid}_{i}"] for i in ccd}
    for i,s_ in fccd.items():
        verts = np.asarray(s_.exterior.xy).T
        ax.add_patch( Polygon(verts, facecolor=facecolor, edgecolor=edgecolor, **kwargs))
        ax.text(*np.mean(verts,axis=0),i, color=textcolor, va="center", ha="center")

    if autoscale:
        ax.autoscale()
        
    return fig

def show_gri_fields(fieldsg=None, fieldsr=None, fieldsi=None,
                    fig=None,
                    title=" ", alignment="horizontal",
                    show_ztf_fields=True, colorbar=True,
                    show_mw=True, mw_b=None, mw_prop={},
                    projection="hammer", moveup=0.05, vscale=1, hscale=1,
                    **kwargs):
    """  """
    prop = {**dict(colorbar=colorbar, edgecolor="0.5", linewidth=0.5),**kwargs}
    
    used_fields = {i+1:f for i,f in enumerate([fieldsg,fieldsr,fieldsi]) if f is not None and len(f)>0}
    
    # None
    if len(used_fields) == 0:
        raise ValueError("No fields given")
    
    # Only one
    if len(used_fields) == 1:
        warnings.warn("Only one color given, favor using show_fields() directly")
        which = list(used_fields.keys())[0]
        return show_fields(used_fields[which], cmap=FIELD_CMAP[which],
                            show_ztf_fields=show_ztf_fields,
                            show_mw=show_mw, mw_b=mw_b, mw_prop=mw_prop,
                            **prop)
    
    # 2 or More
    onlytwo = len(used_fields)==2
    ax, cax = _get_gri_axes_(alignment=alignment, title=title, projection=projection,
                             moveup=moveup, onlytwo=onlytwo, fig=fig, vscale=vscale, hscale=hscale)
    
    fbands = list(used_fields.keys())
    afields_ = list(used_fields.values())
    for i,ax_,cax_,fields_ in zip(fbands, ax, cax, afields_):
        
        if fields_ is not None:
            patch = {}
            if len(np.unique(fields_))==1:
                patch["inclhist"] = False
                if np.unique(fields_)[0]==1:
                    patch["colorbar"] = False
                
            _ = show_fields(fields_, ax=ax_, cax=cax_, cmap=FIELD_CMAP[i],
                            show_ztf_fields=show_ztf_fields,
                            show_mw=show_mw, mw_b=mw_b, mw_prop=mw_prop,
                            **{**prop,**patch})
    return _

def _get_gri_axes_(alignment="classic", title=None, titlefontsize="large", projection="hammer",
                   labelsize="x-small", labelcolor="0.7",  clabelsize="x-small", clabelcolor="k",
                   moveup=None, fig=None, vscale=1, hscale=1, onlytwo=False):
    """ """
    if alignment is None or onlytwo:
        alignment = "flat"

    if moveup is None:
        moveup=0        
        
    if alignment in ["classic"]:
        if fig is None:
            fig = mpl.figure(figsize=[9,6])
        # G
        axg   = fig.add_axes([0.03,0.52+moveup,0.43,0.48], projection=projection)
        caxg  = fig.add_axes([0.03,0.54+moveup,0.43,0.015])
        # R
        axr   = fig.add_axes([0.54,0.52+moveup,0.43,0.48], projection=projection)
        caxr  = fig.add_axes([0.54,0.54+moveup,0.43,0.015])
        # I
        axi   = fig.add_axes([0.27,0.04+moveup,0.43,0.48], projection=projection)
        caxi  = fig.add_axes([0.27,0.05+moveup,0.43,0.015])
        ax = [axg,axr,axi]
        cax = [caxg,caxr,caxi]
    elif alignment in ["flat","aligned", "horizontal"]:
        naxes = 3 if not onlytwo else 2
        if fig is None:
            fig = mpl.figure(figsize=[3.2*naxes,2.5])
        # G
        spanx, spanm = 0.05,0.05
        width = (1-(2*spanx+2*spanm))/naxes
        ax  = [fig.add_axes([ spanm+(i*(width*hscale+spanx)), (0.20+moveup)*vscale, width*hscale, 0.700*vscale],
                                projection=projection) for i in range(naxes)]
        cax = [fig.add_axes([ spanm+(i*(width*hscale+spanx)), (0.12+moveup)*vscale, width*hscale, 0.025*vscale])
                   for i in range(naxes)]
    else:
        raise ValueError(f"cannot parse the given show_gri alignment {alignment}, classic or horizontal")

    if title is not None:
        fig.suptitle(title, fontsize=titlefontsize)
    # labels
    for ax_ in ax:
        ax_.tick_params(labelsize=labelsize, labelcolor=labelcolor)
    for ax_ in cax:        
        ax_.tick_params(labelsize=clabelsize, labelcolor=clabelcolor)
        
    return ax,cax


def show_ztf_fieldvalues(key="Ebv", fieldid="main", mindec=-30,
                        vmin=None, vmax=None,
                        ax=None, cmap="viridis", title=None,
                        colorbar=True, cax=None, clabel=" ",
                        show_ztf_fields=True, grid="main", grid_prop={},
                        show_mw=True, mw_b=None, mw_prop={},
                        savefile=None, **kwargs):
    """ """
    if key not in FIELD_DATAFRAME.columns:
        raise ValueError(f"cannot parse the given key {key}, only columns from FIELD_DATAFRAME available")
    
    
    if type(fieldid) is str or fieldid is None:
        fieldid = get_grid_field(fieldid)
    query_ = "index in @fieldid"
    if mindec is not None:
        query_ +=" and Dec > @mindec"
    # Serie to plot 
    serie = FIELD_DATAFRAME.query(query_)[key]
    return show_fields(fields=serie, vmin=vmin, vmax=vmax,
                           ax=ax, cmap=cmap, title=title,
                           colorbar=colorbar, cax=cax, clabel=clabel,
                           show_ztf_fields=show_ztf_fields, grid=grid, grid_prop=grid_prop,
                           show_mw=show_mw, mw_b=mw_b, mw_prop=mw_prop,
                           savefile=savefile, **kwargs)
    
    
def display_field(ax, fieldid, origin=None, facecolor="0.8", lower_dec=None, edgecolor=None, **kwargs):
    """ """
    print("display_field is DEPRECATED")

def _radec_to_plot_(self, ra, dec):
    """ """
    return np.asarray([-(np.asarray(ra)-self.origin)*np.pi/180, np.asarray(dec)*np.pi/180])

        
##############################
#                            #
#  Individual Field Class    #
#                            #
##############################
class FieldPlotter( object ):
    """ """
    def __init__(self, ax=None, origin=180, inclcax=True, inclhist=False, cax=None, hcax=None, **kwargs):
        """ """
        self.origin = origin        
        self.load_ax(ax, inclcax=inclcax, cax=cax, inclhist=inclhist, **kwargs)

        
    def load_ax(self, ax=None, update_ticks=True, cax=None, hcax=None, inclcax=True, inclhist=False, figsize=None,
                    **kwargs):
        """ """
        if ax is None or len(np.atleast_1d(ax))==4:
            self.fig = mpl.figure(figsize=(8,5) if figsize is None else figsize)
            self.ax = self.fig.add_axes([0.15,0.2,0.75,0.75] if ax is None else ax, projection="hammer")
        else:
            self.ax = ax
            self.fig = self.ax
            
        if inclcax:
            from .utils.plots import HistColorbar
            if cax is None and hcax is None:
                from .utils.plots import insert_ax
                cax = insert_ax(self.ax, "bottom",
                                    shrunk=0.98, space=-0.15, axspace=0.13)
            if inclhist and hcax is None:
                if len(np.atleast_1d(cax)) == 4:
                    xmin, ymin, width, height= cax
                    hcax = [xmin, ymin+np.min([height*1.5,height+0.005]), width, height*1.8]
                else:    
                    bcax = cax.get_position()
                    xmin, ymin, width, height= bcax.xmin, bcax.ymin, bcax.width, bcax.height
                    hcax = cax.figure.add_axes([xmin, ymin+np.min([height*1.5,height+0.005]) , width, height*1.8])
                    
            self.histcbar = HistColorbar(ax=hcax, cax=cax, fig=self.fig, draw=False)
        else:
            self.histcbar = None
            
        if update_ticks:
            tick_labels = np.array([150, 120, 90, 60, 30, 0, 330, 300, 270, 240, 210])
            tick_labels = np.remainder(tick_labels+360+self.origin,360)
            self.ax.set_xticklabels(tick_labels)     # we add the scale on the x axis

        # Label param
        if len(kwargs)>0:
            self.ax.tick_params(**kwargs)
            
    # ---------- #
    #  PLOTTER   #
    # ---------- #
    def show_ztf_grid(self, which="main", 
                  facecolor="None", edgecolor="0.7", alpha=0.1, zorder=1, **kwargs):
        """ """
        fields_ = get_grid_field(which)
        self._ztfgrid = self.show_fields(fields_, 
                                             facecolor=facecolor, edgecolor=edgecolor, 
                                             alpha=alpha, zorder=zorder, **kwargs)

    def show_milkyway(self, b=None, nbins=100, l_start=-241, l_stop=116,**kwargs):
        """ """
        
        nbins=100
        if b is None:
            gal = coordinates.Galactic(np.linspace(l_start,l_stop,nbins)*units.deg, np.zeros(nbins)*units.deg
                                           ).transform_to(coordinates.ICRS)
            prop = dict(ls="-", color="0.7", alpha=0.5)
            self.ax.plot(*self.radec_to_plot(gal.ra, gal.dec), **{**prop, **kwargs})
        else:
            gal_dw = coordinates.Galactic(np.linspace(l_start,l_stop,100)*units.deg, +b*np.ones(nbins)*units.deg
                                         ).transform_to(coordinates.ICRS)
            gal_up = coordinates.Galactic(np.linspace(l_start,l_stop,100)*units.deg, -b*np.ones(nbins)*units.deg
                                         ).transform_to(coordinates.ICRS)
            ra_dw,dec_dw = self.radec_to_plot(gal_dw.ra, gal_dw.dec)
            ra_up,dec_up = self.radec_to_plot(gal_up.ra, gal_up.dec)

            prop = dict(facecolor="0.7", alpha=0.2)
            self.ax.fill_between(ra_dw, dec_dw, dec_up, **{**prop, **kwargs})
        
    def add_fields(self, fields, facecolor="0.7", edgecolor="k", lw=0.5, **kwargs):
        """ """
        fields_verts = self.get_field_vertices(fields)
        self.poly_ = [self.ax.add_patch(Polygon(p_, facecolor=facecolor, 
                                        edgecolor=edgecolor, lw=lw, **kwargs))
                     for p_ in fields_verts]

    def show_fields(self, fields, cmap=None,
                        colorbar=True, clabel=None, cfontsize=None,
                        vmin=None, vmax=None, bins="auto", **kwargs):
        """ fields could be a list of field or a dictionary with single values """
        
        # For now, then will move to pandas.Series as native
        if type(fields) is pandas.Series:
            fields = fields.to_dict()
            
        if type(fields)==dict:
            # popint out nans
            fields = {f:v for f,v in fields.items() if not np.isnan(v)} 
            values = list(fields.values())
            if len(values)==0 or not np.any(values):
                if self.histcbar is not None:
                    self.histcbar.set_visible(False)
                return
            
            if vmin is None: vmin = "0"
            if type(vmin) == str:
                vmin=np.percentile(values, float(vmin))
            if vmax is None: vmax = "100"
            if type(vmax) == str:
                vmax=np.percentile(values, float(vmax))

            values = np.asarray(values)
            if self.histcbar is not None:
                if bins is None or bins in ['auto']:
                    if values.dtype == int:
                        bins = int((vmax-vmin))+1
                        vmax +=1
                self.histcbar.build_histrogram(data=values, vmin=vmin, vmax=vmax, bins=bins)
                
            # - CBAR
            if cmap is None and self.histcbar is not None:
                cmap = self.histcbar.cmap
                
            elif type(cmap) == str:
                if self.histcbar is not None:
                    self.histcbar.load_cmap(cmap)
                    cmap = self.histcbar.cmap
                else:
                    cmap = mpl.get_cmap(cmap)
            else:
                self.histcbar.load_cmap(cmap.name)
                
                
#            _ = kwargs.pop("facecolor",None) #remove facecolor is any
            for f,v in fields.items():
                self.add_fields(f, **{**dict(facecolor=cmap((v-vmin)/(vmax-vmin)) if vmax-vmin !=0 else cmap(0)), **kwargs})

            # - Cbar
            if colorbar:
                self.show_colorbar(clabel=clabel, cfontsize=cfontsize)
            elif self.histcbar is not None:
                 self.histcbar.set_visible(False)
                 
        else:
            self.add_fields(fields, **kwargs)
        
    def show_point(self, radec, **kwargs):
        """ """
        xy = self.radec_to_plot(*radec)
        self.ax.scatter(*xy, **kwargs)

    def show_colorbar(self, clabel=None, cfontsize=None):
        """ """
        if self.histcbar is None:
            return

        self.histcbar.draw()
        self.histcbar.set_label(clabel, fontsize=cfontsize)
        
    def get_field_vertices(self, fields_):
        """ Get the field vertices in plotting coordinates. """
        fields_ = np.squeeze(get_field_vertices(fields_, squeeze=False), axis=1) # remove CCDs
        return [self.radec_to_plot(*f_.T).T for f_ in fields_] 
    
    def radec_to_plot(self, ra, dec):
        """ """
        return np.asarray([-(np.asarray(ra)-self.origin)*np.pi/180, np.asarray(dec)*np.pi/180])


   
##############################
#                            #
#    ZTF Fields Class        #
#                            #
##############################

class FieldAnimation( FieldPlotter ):
    
    def __init__(self, fields, ax=None,dates=None, facecolors=None, alphas=None, edgecolors=None, inclcax=False):
        """ """
        super().__init__(ax=ax, inclcax=inclcax)
        
        self.set_fields(fields)
        self.set_dates(dates)
        self.set_properties(facecolors=facecolors, alphas=alphas, edgecolors=edgecolors)
        
    # ================= #
    #   Methods         #
    # ================= #
    def set_dates(self, dates):
        """ """
        self._dates = np.atleast_1d(dates) if dates is not None else None

    def set_fields(self, fields):
        """ """
        self._fields = fields
        self._unique_fields = np.unique(self._fields)
        self._field_vertices = {i:v for i,v in zip(self._unique_fields,self.get_field_vertices(self._unique_fields))}
        
    def set_properties(self, facecolors=None, edgecolors=None, alphas=None):
        """ """
        self._display_prop = {}
        self._set_prop_("facecolor", facecolors, "0.7")
        self._set_prop_("edgecolor", edgecolors, "None")
        self._set_prop_("alpha", alphas, 1)

    def _set_prop_(self, key, value, default=None):
        """ """
        if not hasattr(self,"_display_prop"):
            self._display_prop = {}
            
        if value is None:
            value = default
            
        if len(np.atleast_1d(value)) == 1:
            self.display_prop[key] = np.atleast_1d(value)
            self.display_prop[f"unique_{key}"] = True
        else:
            self.display_prop[key] = value
            self.display_prop[f"unique_{key}"] = False
            
    # ---------- #
    #  SETUP     #
    # ---------- #
    def reset(self):
        """ """
        self.intpoly_ = Polygon(self.field_vertices[self.fields[0]],
                                    facecolor=self.display_prop["facecolor"][0],
                                    edgecolor=self.display_prop["edgecolor"][0],
                                    alpha=self.display_prop["alpha"][0])
        
        if self._dates is not None:
            self.inttext_ = self.fig.text(0.01,0.99, self._dates[0],
                                           va="top", ha="left", weight="bold")
            
        p_ = self.ax.add_patch(self.intpoly_)
        return self.intpoly_
        

    # ---------- #
    #  Animate   #
    # ---------- #    
    def update_field_to(self, i):
        """ """
        try:
            self.intpoly_.set_xy(self.field_vertices[ self.fields[i] ])
            if self.dates is not None and len(self.dates)>1:
                self.inttext_.set_text(self.dates[i])
        except:
            print(f"FAILES for i={i}")
            
        for key in ["facecolor","edgecolor","alpha"]:
            if not self.display_prop[f"unique_{key}"]:
                getattr(self.intpoly_,f"set_{key}")(self.display_prop[key][i])
                
        return self.intpoly_

    def launch(self, interval=5, repeat=False, blit=True, savefile=None):
        """ """
        from matplotlib import animation
        self.anim = animation.FuncAnimation(self.fig, self.update_field_to,
                                                init_func=self.reset,
                               frames=self.nfields, interval=interval, repeat=repeat, blit=blit)
        
    # ================= #
    #   Properties      #
    # ================= #
    @property
    def fields(self):
        """ Fields that should be shown """
        return self._fields

    @property
    def dates(self):
        """ Observation dates if any """
        return self._dates
    @property
    def nfields(self):
        """ size of self.fields """
        return len(self.fields)

    @property
    def field_vertices(self):
        """ vertices of the fields """
        return self._field_vertices

    @property
    def display_prop(self):
        """ """
        return self._display_prop


    
##############################
#                            #
#    Planner Class           #
#                            #
##############################

class PalomarPlanning( object):
    """ """
    def __init__(self, date=None, **kwargs):
        """ """
        self._site = coordinates.EarthLocation.of_site("palomar")
        self._utcshift = -8*units.h
        if date is not None:
            self.set_date(date, **kwargs)

    # --------- #
    #  SETTER   #
    # --------- #
    def set_date(self, date, timerange=[-7,7], to_utc=True, **kwargs):
        """ """
        self._date = date
        self.set_night(date,timerange=[-7,7], to_utc=True, **kwargs)
        
    def set_night(self, night, timerange=[-7,7], to_utc=True, **kwargs):
        """ """    
        self._night = self.get_night(night, timerange=timerange, to_utc=to_utc, **kwargs)
        self._nightaltaz = self._get_night_altaz_(self.night)
        
    # --------- #
    #  GETTER   #
    # --------- #
    @classmethod
    def get_date_night_duration(cls, date, twilight=-12*units.deg, **kwargs):
        """ ClassMethod to directly get the night duration at a given date or list of dates. """
        return cls().get_night_duration(date, twilight=twilight, **kwargs)

    
    def get_night(self, date, timerange=[-7,7], to_utc=True, range_units=units.h, bins=100, **kwargs):
        """ """
        if type(date) is not time.Time:
            date = time.Time(date, **kwargs)
        if to_utc:
            date -= self.utcshift
            
        if timerange is None:
            return date
        if len(np.atleast_1d(date))==1:
            return date + np.linspace(*timerange, bins)*range_units
        return date + np.linspace(*timerange, bins)[:,None]*range_units

    def get_night_duration(self, date=None, twilight=-12*units.deg, bins=500):
        """ in hours """
        if date is None:
            date =self.date
        
        time_, sunaltaz = self.get_body_altaz("sun",  date, bins=bins)
        flagnight = sunaltaz.alt<twilight
        
        if len(np.atleast_1d(date))==1:
            nigh_time = time_[flagnight][[0,-1]].jd
            return np.diff(nigh_time)[0]*units.day.to("h")*units.h
        
        length = []
        for i,d_ in enumerate(date):
            nigh_time_ = time_[:,i][flagnight[:,i]][[0,-1]].jd
            night_length = np.diff(nigh_time_)[0]*units.day.to("h")*units.h
            length.append(night_length)
        return length

    def get_fields_altaz(self, fieldid, date=None, **kwargs):
        """ """
        radec = get_field_centroid(fieldid)
        return self.get_coord_altaz(radec, date=date, **kwargs)

    def get_coord_altaz(self, radec, date=None, **kwargs):
        """ """
        if type(radec) is not coordinates.SkyCoord:
            radec = coordinates.SkyCoord(radec, unit="deg")
            
        date, datealtaz = self._read_date_input_(date, **kwargs)
            
        return date, radec.transform_to( datealtaz[:,None] )
    
    def get_body_altaz(self, bodyname, date=None, **kwargs):
        """ 
        'sun', 'moon', 'mercury', 'venus', 'earth-moon-barycenter', 'mars', 'jupiter', 'saturn', 'uranus', 'neptune'
        """
        date, datealtaz = self._read_date_input_(date, **kwargs)
        body = coordinates.get_body(bodyname, date)
        return date, body.transform_to( datealtaz )
        
        
    def get_fields_observability(self, fieldid, date=None, airmasslimit=[1,1.5], minobservability=90*units.min,
                                     twilight=-18*units.deg, still_observablein=None):#2*units.week):
        """ """
        nighttime, field_altaz = self.get_fields_altaz(fieldid, date=date)

        airmasses = np.asarray(field_altaz.secz)
        # Twilight Cut
        flagtwilight = self.is_twilight(date, twilight, squeeze=True)
        flag_good = ((~flagtwilight[:,None]) * (airmasses>=airmasslimit[0]) * (airmasses<airmasslimit[1]) )
        if minobservability is not None:
            min_night_frac = minobservability/(nighttime[~flagtwilight].max()-nighttime[~flagtwilight].min()).to("min")
            flag_nightfrac = (np.sum(flag_good, axis=0)/len(nighttime[~flagtwilight])>=min_night_frac)
            flag_good *= flag_nightfrac[None,:]


        stime, ffrac = [pandas.Series(np.sum(flag_good, axis=1), index=pandas.DatetimeIndex(nighttime.datetime)),
                        pandas.Series(np.sum(flag_good, axis=0)/len(nighttime[~flagtwilight]), index=fieldid)]
        
        if still_observablein is not None:
            print("using still_observablein options") 
            if date is None:
                date = self.date
                
            delay_date = (time.Time(date) + still_observablein).iso.split(" ")[0]
            delay_stime, delay_ffrac = self.get_fields_observability(fieldid, date=delay_date,
                                                                    airmasslimit=airmasslimit, minobservability=minobservability,
                                                                    twilight=twilight, still_observablein=None)
            fieldidstillin = ffrac[ffrac>0].index[np.in1d(ffrac[ffrac>0].index,delay_ffrac[delay_ffrac>0].index)]
            flagstillin = np.in1d(ffrac.index,fieldidstillin)
            flag_good *= flagstillin[None,:]
            
        return [pandas.Series(np.sum(flag_good, axis=1), index=pandas.DatetimeIndex(nighttime.datetime)),
                pandas.Series(np.sum(flag_good, axis=0)/len(nighttime[~flagtwilight]), index=fieldid)]
    
        
    def get_observable_fields(self, fieldids, date=None, airmasslimit=[1,1.5],
                                minobservability=90*units.min, twilight=-18*units.deg,
                                **kwargs):
        """ """
        stime, ffrac = self.get_fields_observability(fieldids, date=date, airmasslimit=airmasslimit,
                                                     twilight=twilight, minobservability=minobservability,
                                                     **kwargs)
        return ffrac[ffrac>0].index
    
        
    def is_twilight(self, date=None, sunlimit=-18*units.deg, squeeze=True):
        """ """
        time_, sunaltaz = self.get_body_altaz("sun", date)
        sunlimit = np.atleast_1d(sunlimit)
        
        flags = [sunaltaz.alt>slimit for slimit in sunlimit]
        
        if len(sunlimit) ==1 and squeeze:
            return flags[0]
        return flags
    
    def is_day(self, date, sunlimit=0*units.deg, squeeze=True, **kwargs):
        """ """
        return self.is_twilight(sunlimit=0*units.deg, squeeze=squeeze, **kwargs)

    def _get_night_altaz_(self, night, **kwargs):
        """ 
        **kwargs goes to astropy.time.Time if date is not one already.
        """
        return coordinates.AltAz(obstime=night, location=self.site)
    
    def _read_date_input_(self, date, **kwargs):
        """ """
        if date is not None:
            if type(date) is not time.Time:
                date = self.get_night(date, **kwargs)
            datealtaz = self._get_night_altaz_(date)
        elif self.has_night():
            date      = self.night
            datealtaz = self.nightaltaz
        else:
            raise ValueError("No night set (self.set_night), no date given (see date option).")
            
        return date, datealtaz
    # -------- #
    #  PLOTTER #
    # -------- #
    def show(self, date=None, fields=None, radec=None, body=None, propdate={},
            show_twilight=True, as_airmass=False, ctwilight="0.7"):
        """ """
        import matplotlib.pyplot as mpl
        from matplotlib import dates as mdates
        
        fig = mpl.figure(figsize=[7,4])
        ax = fig.add_subplot(111)

        if date is not None:
            if type(date) is not time.Time:
                date = self.get_night(date, **propdate)
        else:
            date = self.night
            
        # 
        # - start: Plotting
        for type_ in ["fields", "radec", "body"]:
            t_ = eval(type_)
            if t_ is not None:
                _, v_ = getattr(self,f"get_{type_}_altaz")(t_, date)
                ax.plot(date.datetime, v_.alt if not as_airmass else v_.secz)
        # - end: Plotting
        # 
            
        # 
        # - start: Twilight bands
        if show_twilight:
            self.show_twilight(ax=ax, date=date)
            
        # - start: End bands
        # 
        
        # - 
        ax.set_ylabel("Altitude [deg]" if not as_airmass else "Airmass")
        # - cleaning
        locator = mdates.AutoDateLocator()
        formatter = mdates.ConciseDateFormatter(locator)
        ax.xaxis.set_major_locator(locator)
        ax.xaxis.set_major_formatter(formatter)
        # - Limites
        ax.set_xlim(date.datetime[0],date.datetime[-1])
        
        ax.set_ylim(-1,3) if as_airmass else ax.set_ylim(-10,95)
        
    def show_twilight(self, date=None, ax=None, sunlimit=[0,-12,-18]*units.deg,
                     ctwilights="0.7", alphas=[0.8,0.3,0.1], propdate={}):
        """ """
        import matplotlib.pyplot as mpl
                
        ctwilights = np.atleast_1d(ctwilights)
        alphas = np.atleast_1d(alphas)
        
        if len(ctwilights)==1:
            ctwilights = list(ctwilights)*len(sunlimit)
        if len(alphas)==1:
            alphas = list(alphas)*len(alphas)
            
        if ax is None:
            fig = mpl.figure(figsize=[7,4])
            ax = fig.add_subplot(111)
        else:
            fig = ax.figure
            
        
        if date is not None:
            if type(date) is not time.Time:
                date = self.get_night(date, **propdate)
        else:
            date = self.night
        
        for i, tflag in enumerate(self.is_twilight(date=date, sunlimit=sunlimit, squeeze=False)):
            ax.fill_between(date.datetime, 0,1, where=tflag, 
                            transform=ax.get_xaxis_transform(), 
                            color=ctwilights[i], alpha=alphas[i])
        
        
    def show_fields_observability(self, fieldids, show_twilight=True, cmap="viridis",
                                      airmasslimit=[1,1.5], twilight=-18*units.deg, minobservability=90*units.min,
                                      show_notobs=True, cmapv=0.1,
                                      fcno="0.7",ecno="0.7", alphano=0.15, nolw=0, noprop={},
                                      **kwargs):
        """ 
        **kwargs foes to get_fields_observability() 
        """
        import matplotlib.pyplot as mpl        
        from matplotlib import dates as mdates

        figsize = [9,3.5]
        axmap   = [0.05,0.21,0.45,0.7]
        caxmap  = [0.05,0.18,0.45,0.02]

        stime, ffrac = self.get_fields_observability(fieldids, airmasslimit=airmasslimit, twilight=twilight, minobservability=minobservability)
        
        fig = show_fields( ffrac[ffrac.values>0],
                               bkgd_fields=fieldids[ffrac.values==0] if show_notobs else None,
                               bkgd_prop={**dict(facecolor=fcno, edgecolor=ecno, alpha=alphano, lw=nolw),**noprop},
                                  ax=axmap, cax=caxmap, figsize=figsize, cmap=cmap,
                                  axparam={"labelsize":"x-small","color":"0.7", "labelcolor":"0.7"},
                                  clabel="Fraction of night observable", cfontsize="medium")

        axh = fig.add_axes([0.6,0.17,0.35,0.68])
        axh.fill_between(stime.index, stime.values, 
                         facecolor=mpl.cm.get_cmap(cmap)(cmapv,0.2), 
                         edgecolor=mpl.cm.get_cmap(cmap)(cmapv,0.9), lw=2, **kwargs)


        locator = mdates.AutoDateLocator(minticks=4, maxticks=6)
        formatter = mdates.ConciseDateFormatter(locator)
        axh.xaxis.set_major_locator(locator)
        axh.xaxis.set_major_formatter(formatter)
        axh.set_ylim(bottom=0)
        axh.set_xlim(*self.night.datetime[[0,-1]])
        axh.set_ylabel("Number of fields observable")
        if show_twilight:
            self.show_twilight(ax=axh, date=self.night, ctwilights="0.9")
            
        fig.text(0.02,0.98, f"Field Observability | {len(ffrac[ffrac>0])} fields", color="0.7", fontsize="medium", va="top",ha="left")
        
        return fig
    
    # ============= #
    #  Properties   #
    # ============= #
    @property
    def site(self):
        """ """
        return self._site
    @property
    def utcshift(self):
        """ """
        return self._utcshift

    @property
    def date(self):
        """ """
        return None if not hasattr(self,"_date") else self._date
    
    @property
    def night(self):
        """ """
        if not self.has_night():
            return None
        return self._night
    
    def has_night(self):
        """ """
        return getattr(self,"_night") and self._night is not None
    
    @property
    def nightaltaz(self):
        """ """
        if not self.has_night():
            raise AttributeError("No night set.")
            
        return self._nightaltaz



""" High level tools design for cluster computation """

import numpy as np
import warnings
# Cluster tool
import dask

class ClusterQuery( object ):
    """ """
    def __init__(self, client=None):
        """ """
        if client is not None:
            self.set_client(client)

    # ============= #
    #  Methods      #
    # ============= #
    # ---------- #
    #  Clients   #
    # ---------- #
    def set_client(self, client):
        """ """
        self._client = client

    def get_client(self, client=None):
        """ returns the given client if any, otherwise the current loaded client. 
        None if nothing 
        """
        if client is not None:
            return client
        
        if self.has_client():
            return self.client
        
        warnings.warn("No dask client given and no dask client sent to the instance. None returned")
        return None

    # ---------- #
    #  Tools     #
    # ---------- #
    def ctest_fits_files(self, files, client=None, chunks=300, test_exist=False):
        """ """
        from .io import _are_fitsfiles_bad_
        if len(files)<chunks:
            raise ValueError(f"more chunks then files: {len(files)} files of {chunks} chunks")
        
        chunked_files = np.array_split(files, chunks)
        d_test = [dask.delayed(_are_fitsfiles_bad_)(files_, test_exist=test_exist)
                      for files_ in chunked_files]
        
        client_ = self.get_client(client)
        if client_ is None:
            return d_test
        
        futures = client_.compute(d_test)
        bools = client_.gather(futures)
        return np.asarray([chunked_files[i][b] for i,b in enumerate(bools) if np.any(b)])

    def cget_files(self, files, client=None, chunks=300, **kwargs):
        """ """
        def get_files(files_, **kwargs):
            return [io.get_file(f_, show_progress=False, maxnprocess=1, **kwargs)
                        for f_ in files_]
        
        if len(files)<chunks:
            raise ValueError(f"more chunks then files: {len(files)} files of {chunks} chunks")
        
        chunked_files = np.array_split(files, chunks)
        d_get = [dask.delayed(get_files)(files_) for files_ in chunked_files]
        
        client_ = self.get_client(client)
        if client_ is None:
            return d_get

        return client_.compute(d_test)
        
    # ============= #
    #  Properties   #
    # ============= #
    @property
    def client(self):
        """ Client used to run the methods on cluster """
        return self._client

    def has_client(self):
        """ has a client been set ? """
        return hasattr(self,"_client") and self._client is not None
        



#!/usr/bin/env python
#


import pandas
BTS_URL = "https://sites.astro.caltech.edu/ztf/bts/explorer.php?"
_PRIVATE_BTS_URL = "http://sites.astro.caltech.edu/ztf/rcf/explorer.php?"

def download_bts_table(subsample="cantrans", quality="y", endpeakmag=19.0,
                       classstring=None, classexclude=None,  ztflink=None,
                       lastdet=None, startsavedate=None, startpeakdate=None, 
                       startra=None, startdec=None, startz=None, startdur=None, 
                       startrise=None, startfade=None, startpeakmag=None, startabsmag=None,
                       starthostabs=None, starthostcol=None, 
                       startb=None, startav=None, endsavedate=None, endpeakdate=None,
                       endra=None, enddec=None, endz=None, enddur=None, endrise=None, endfade=None,
                       endabsmag=None, endhostabs=None, endhostcol=None, endb=None, endav=None):
    """ Get the current ZTF Bright Transient Survey (BTS) list.
    
    
    Credits:
    --------
    Please cite: 
    - Perley et al. 2020, ApJ, 904, 35
      https://arxiv.org/abs/2009.01242
    and 
    - Fremling et al. 2020, ApJ , 895, 32
      https://arxiv.org/abs/1910.12973

    Parameters
    ----------
    See documentation here https://sites.astro.caltech.edu/ztf/bts/explorer_info.html

    Returns
    -------
    DataFrame
    """
    query = [f"{k}={v}" for k,v in locals().items() if v is not None]
    
    url = BTS_URL+"f=s&"+"&".join(query)+"&format=csv"
    #print(url)
    data = pandas.read_csv(url).set_index("ZTFID")
    data["peakt_jd"] = data["peakt"].astype("float") + 2458000
    return data

def download_bts_privatetable( auth=None,
                                  sort="savedate", subsample="cantrans",endpeakmag=19.0,
                                  quality="y", ztflink="marshal",
                                  coverage=None, 
                                  classstring=None,classexclude=None,
                                  startsavedate=None, startpeakdate=None, startlastdate=None,
                                  startra=None, startdec=None, startz=None,
                                  startdur=None, startrise=None, startfade=None, startpeakmag=None,
                                  startlastmag=None, startabsmag=None, starthostabs=None,
                                  starthostcol=None, startsavevis=None, startlatevis=None,
                                  startcurrvis=None, startb=None,
                                  startav=None, endsavedate=None, endpeakdate=None, endlastdate=None,
                                  endra=None, enddec=None, endz=None, enddur=None, endrise=None,
                                  endfade=None, endlastmag=None, endabsmag=None, endhostabs=None, endhostcol=None,
                                  endsavevis=None, endlatevis=None, endcurrvis=None, endb=None, endav=None):
    """ Get the current ZTF Bright Transient Survey (BTS) list.
    
    Credits:
    --------
    Please cite: 
    - Perley et al. 2020, ApJ, 904, 35
      https://arxiv.org/abs/2009.01242
    and
    - Fremling et al. 2020, ApJ , 895, 32
      https://arxiv.org/abs/1910.12973

    Parameters
    ----------
    See documentation here https://sites.astro.caltech.edu/ztf/bts/explorer_info.html

    Returns
    -------
    DataFrame
    """
    query = [f"{k}={v}" for k,v in locals().items() if v is not None]

    import requests
    from . import io
    
    url = _PRIVATE_BTS_URL+"f=s&"+"&".join(query)+"&format=csv"
    r = requests.get(url, auth= io._load_id_("bts") if auth is None else auth)
    columns, *data = [l.split(",") for l in r.text.split("\n") if len(l)>5]
    data = pandas.DataFrame(data, columns=columns).set_index("ZTFID")
    data["peakt_jd"] = data["peakt"].astype("float") + 2458000
    return data

    





#!/usr/bin/env python
#

""" Alert Reader and basic tools """


import numpy as np

FILTER_COLORS = ["C2","C3","C1"]
FILTER_CODE   = ["zg","zr","zi"]


def display_alert(alert, savefile=None, show_ps_stamp=False):
    """ """
    if type(alert) == str:
        # This is a file:
        al = AlertReader.load(alert)
    else:
        al = AlertReader(alert)

    return al.show(savefile, show_ps_stamp=show_ps_stamp)



def query_alert(alertid):
    """ """
    print("This service will soon be added")
    pass

####################
#                  #
#  Basic Class     #
#                  #
####################
class AlertReader():
    """ """    
    def __init__(self, alert):
        """ Input: Avro alert loaded. If you have a avrofile, use .load()"""
        self.alert = alert

        
    @classmethod
    def load(cls, avrofile):
        """ """
        import fastavro
        with open(avrofile, "rb") as fo:
            av = next(fastavro.reader(fo), None)
        return cls(av)
    

    # ================ #
    #   Methods        #
    # ================ #
    def get_stamp(self, which):
        """ return the HDUImage stamps for `which`.
        where `which` is one of the following ['Science','Template','Difference']
        """
        import gzip, io
        from astropy.io import fits
        with gzip.open(io.BytesIO(self.alert.get("cutout%s"%which)["stampData"]), 'rb') as f:
            return fits.open(io.BytesIO(f.read()))[0]

    # --------- #
    #  GETTER   #
    # --------- #
    def get_history_photopoints(self):
        """ list of alert history having been detected (not upper limit, real photometric points) """
        return [d for d in self.alert['prv_candidates'] if d.get('candid') is not None]

    def get_history_upperlimits(self):
        """ list of alert history not  detected (upper limit, not real photometric points) """
        return [d for d in self.alert['prv_candidates'] if d.get('candid') is None]

    def download_ps_stamp(self, size=240, color=["y","g","i"]):
        """ Downloading the PS colored cutout stamp centered at the alert candidate location
        Returns the PIL Image
         = uses `get_ps_stamp` from ztfquery.utils.stamps.
        """
        from .utils.stamps import get_ps_stamp
        return get_ps_stamp(self.alert["candidate"]["ra"], self.alert["candidate"]["dec"],  size=size, color=color)
    
    # --------- #
    # PLOTTER   #
    # --------- #
    def show(self, savefile=None, show_ps_stamp=False, pcvmin=0, pcvmax=100):
        """ 
        Parameters
        ----------
        pcvmin, pcvmax: [0<float<100] -optional-
            percentile of the saturating (min and max) color for the stamps
        """
        from astropy import visualization
        from astropy.time import Time
        import matplotlib.dates as mdates
        from matplotlib.colors import Normalize
        import matplotlib.pyplot as mpl
    
        # ----------- #
        # Global      #
        # ----------- #
        prop = dict(marker="o", mec="0.7", ms=8, ecolor="0.7", ls="None")
    
        # ----------- #
        #   Methods   #
        # ----------- #
        def show_fid_lc(ax):
            """ """
            if len(self.get_history_photopoints()) == 0:
                return
            mag,magerr, jd, fid = np.asarray([ [d[k] for k in ["magpsf","sigmapsf","jd","fid"]]
                                                   for d in self.get_history_photopoints()]).T
            for j,i in enumerate([1,2,3]):
                if i in fid:
                    flag_fid = fid==i
                    ax.errorbar([Time(jd_, format="jd").datetime for jd_ in jd[flag_fid]], 
                                    mag[flag_fid], yerr= magerr[flag_fid], 
                                    label="magpsf %s"%FILTER_CODE[j], mfc=FILTER_COLORS[j], **prop)
            
        def show_fid_uplim(ax):
            """ """
            if len(self.get_history_upperlimits()) == 0:
                return
            upmag, jdup, fidup    = np.asarray([ [d[k] for k in ["diffmaglim","jd","fid"]]
                                                     for d in self.get_history_upperlimits()]).T
            for j,i in enumerate([1,2,3]):
                if i in fidup:
                    flag_fid = fidup==i
                    ax.errorbar([Time(jd_, format="jd").datetime for jd_ in jdup[flag_fid]], 
                                upmag[flag_fid], yerr=0.2, lolims=True,
                                    color=FILTER_COLORS[j], ls="None", alpha=0.7,
                                    label="_no_legend_")
                
        # ----------- #
        #   Axes      #
        # ----------- #
        fig = mpl.figure(figsize=[9,5])

        ref, width, heigh = 0.1,0.15, 0.25
        ypos, span  = 0.65, 0.05
        # Stamps
        aximg = fig.add_axes([ref,ypos,width,heigh])
        axref = fig.add_axes([ref+(width+span),ypos,width,heigh])
        axdif = fig.add_axes([ref+(width+span)*2,ypos,width,heigh])
        if show_ps_stamp:
            axps   = fig.add_axes([ref+(width+span*1.5)*3,ypos,width,heigh])
            axps.set_yticks([])
            axps.set_xticks([])
        # - LC
        axlc   = fig.add_axes([ref, 0.1,(width+span)*2+width,0.5])
        axlc.set_xlabel("Date", fontsize="large")
        axlc.set_ylabel("mag (magpsf)", fontsize="large")
    
        # ----------------- #
        #  Plotting Stamps  #
        # ----------------- # 
        # Loop Over the stamps
        for ax_, source in zip([aximg, axref, axdif], ['Science','Template','Difference']):
            origdata = self.get_stamp(source).data
            vmin,vmax = np.percentile(origdata[origdata==origdata], [pcvmin,pcvmax])
            data_ = visualization.AsinhStretch()((origdata-vmin)/(vmax-vmin))

            ax_.imshow(data_, norm=Normalize(*np.percentile(data_[data_==data_], [0.5,99.5])), aspect="auto")
            ax_.set_yticks([])
            ax_.set_xticks([])
            ax_.set_title(source)

        if show_ps_stamp:
            try:
                img = self.download_ps_stamp(color=["y","g","i"])
                axps.imshow(np.asarray(img))
            except:
                print("Pan-STARRS stamp failed.")
            axps.set_title("Pan-STARRS (y/g/i)")
        # ----------- #
        #  History    #
        # ----------- #
        show_fid_lc(axlc)
        axlc.invert_yaxis()
        show_fid_uplim(axlc)
        axlc.legend(loc="best")
    
        # ----------- #
        #  Alert      #
        # ----------- #
        prop['marker'] = "D"
        prop['mec'] = FILTER_COLORS[self.alert["candidate"]["fid"]-1]
        axlc.errorbar(Time(self.alert["candidate"]["jd"], format="jd").datetime, self.alert["candidate"]["magpsf"], 
                    yerr= self.alert["candidate"]["sigmapsf"],
                    mfc=FILTER_COLORS[self.alert["candidate"]["fid"]-1],label="_no_legend_", **prop)
    
        # add text
        info = []
        for k in ["rb","fwhm","nbad", "elong", "isdiffpos"]:
            try:
                info.append("%s : %.3f"%(k,self.alert["candidate"].get(k)) )
            except:
                info.append("%s : %s"%(k,self.alert["candidate"].get(k)) )
            
        for kk in ["objectidps", "sgscore", "distpsnr","srmag"]:
            for k in [k for k in self.alert["candidate"].keys() if kk in k]:
                info.append("%s : %s"%(k,self.alert["candidate"].get(k)) )

        fig.text(0.68,0.6, " \n".join(info), va="top", fontsize="medium", color="0.4")
        fig.text(0.005,0.995, "alert: ID: %s (RA: %.5f | Dec: %.5f | Filter: %s)"%(self.alert.get("candid"),
                                                                                   self.alert['candidate']['ra'],self.alert['candidate']['dec'],
                                                                                   FILTER_CODE[self.alert['candidate']['fid']-1]),
                     fontsize="medium", color="k", va="top", ha="left")
        
        locator = mdates.AutoDateLocator()
        formatter = mdates.ConciseDateFormatter(locator)
        axlc.xaxis.set_major_locator(locator)
        axlc.xaxis.set_major_formatter(formatter)


        if savefile is not None:
            fig.savefig(savefile, dpi=250)
            
        return fig



#! /usr/bin/env python
#

DESCRIPTION = "access ZTF data from python "
LONG_DESCRIPTION = """ ZTF data from python """

DISTNAME = 'ztfquery'
AUTHOR = 'Mickael Rigault'
MAINTAINER = 'Mickael Rigault' 
MAINTAINER_EMAIL = 'm.rigault@ipnl.in2p3.fr'
URL = 'https://github.com/MickaelRigault/ztfquery'
LICENSE = 'BSD (3-clause)'
DOWNLOAD_URL = 'https://github.com/MickaelRigault/ztfquery/tarball/1.15'
VERSION = '1.15.7'

try:
    from setuptools import setup, find_packages
    _has_setuptools = True
except ImportError:
    from distutils.core import setup
    _has_setuptools = False
    
def check_dependencies():
   install_requires = []
   return install_requires

if __name__ == "__main__":

    install_requires = check_dependencies()

    if _has_setuptools:
        packages = find_packages()
        print(packages)
    else:
        # This should be updated if new submodules are added
        packages = ['ztfquery']
    
        
    setup(name=DISTNAME,
          author=AUTHOR,
          author_email=MAINTAINER_EMAIL,
          maintainer=MAINTAINER,
          maintainer_email=MAINTAINER_EMAIL,
          description=DESCRIPTION,
          long_description=LONG_DESCRIPTION,
          license=LICENSE,
          url=URL,
          version=VERSION,
          download_url=DOWNLOAD_URL,
          install_requires=install_requires,
          packages=packages,
          package_data={'ztfquery': ['data/*']},
          scripts=["bin/irsa_query.py"],
          classifiers=[
              'Intended Audience :: Science/Research',
              'Programming Language :: Python :: 2.7',
              'Programming Language :: Python :: 3.5',
              'License :: OSI Approved :: BSD License',
              'Topic :: Scientific/Engineering :: Astronomy',
              'Operating System :: POSIX',
              'Operating System :: Unix',
              'Operating System :: MacOS'],
      )



#! /usr/bin/env python
#

"""  Access SEDM data from pharos """


PHAROS_BASEURL = "http://pharos.caltech.edu"
import os
import requests
import json
import numpy as np
import pandas
import warnings
from . import io

SEDMLOCAL_BASESOURCE = os.path.join(io.LOCALSOURCE,"sedm")
SEDMLOCALSOURCE = os.path.join(SEDMLOCAL_BASESOURCE,"redux")
if not os.path.exists(SEDMLOCAL_BASESOURCE):
    os.makedirs(SEDMLOCAL_BASESOURCE, exist_ok=True)
if not os.path.exists(SEDMLOCALSOURCE):
    os.makedirs(SEDMLOCALSOURCE, exist_ok=True)
    

#######################
#                     #
#  High level method  #
#                     #
#######################

#
# - whatfiles
#
def download_whatfile(date, as_dataframe=True, auth=None, store=True):
    """ Download summary of files observed on the given date
    
    Parameters
    ----------
    date: [string]
        Date with the following format: YYYYMMDD

    as_dataframe: [bool] -optional-
        output format: text (False) or DataFrame (True)
    """
    whatfile = _download_sedm_data_(date, "what.list", auth=auth).text.splitlines()
    
    if as_dataframe or store:
        data = whatfiles_to_dataframe(whatfile)
        if store:
            fileout = get_whatfile_path(date)
            if not os.path.isdir( os.path.dirname(fileout) ):
                os.makedirs(os.path.dirname(fileout), exist_ok=True)
                
            data.to_parquet(fileout)
            
    return whatfile if not as_dataframe else data

def download_pharosfile(date, auth=None, store=True):
    """ """

    pharosfile = _download_sedm_pharosfile_(date, auth=auth)
    
    if store:
        fileout = get_pharosfile_path(date)
        if not os.path.isdir( os.path.dirname(fileout) ):
            os.makedirs(os.path.dirname(fileout), exist_ok=True)

        with open(fileout, "w") as fileout_:
            for l_ in pharosfile:
                fileout_.write(l_+"\n")
            
    return pharosfile


def bulk_download(which, dates, client=None):
    """ """
    if which not in ["pharosfile", "whatfile"]:
        raise ValueError(f"which can either be 'pharosfile' or 'whatfile', {which} given")

    download_func = eval(f"download_{which}")
    if client:
        from dask import delayed
        d_download = [delayed(download_func)(date_) for date_ in dates]
        return client.compute(d_download)

    return [download_func(date_) for date_ in dates]




def get_whatfile_path(date, extension=None):
    """ """
    if extension is None:
        extension = ".parquet"
    if not extension.startswith("."):
        extension = f".{extension}"

    if date is None or date == "stored":
        return os.path.join(SEDMLOCAL_BASESOURCE, "whatfiles", f"stored_data{extension}")
    return os.path.join(SEDMLOCAL_BASESOURCE, "whatfiles", f"what_{date.lower()}{extension}")

def get_pharosfile_path(date, extension=None):
    """ """
    if extension is None:
        extension = ".dat"
        
    if not extension.startswith("."):
        extension = f".{extension}"
        
    return os.path.join(SEDMLOCAL_BASESOURCE, "pharosfiles", f"pharosfile_{date.lower()}{extension}")

# Internal Low level download

def _download_sedm_data_(night, pharosfile, fileout=None, auth=None,verbose=False):
    """ """
    url = os.path.join(PHAROS_BASEURL,"data",night,pharosfile)
    if verbose:
        print(url)
    if auth is None:
        auth = io._load_id_("pharos")
    return io.download_single_url(url,fileout=fileout,
                                  auth=auth,
                                  cookies="no_cookies")

def _download_sedm_pharosfile_(date, auth=None):
    """ """
    username,password = io._load_id_("pharos") if auth is None else auth
    requests_prop = {"data":json.dumps({"obsdate":date,
                                            "username":username,
                                            "password":password,
                                            }),
                         "headers":{'content-type': 'application/json'}}
            
    t = requests.post(os.path.join(PHAROS_BASEURL,"get_user_observations"), **requests_prop).text
    try:
        t = json.loads(t)
    except:
        raise IOError("Cannot load the downloaded file (not a json) ")

    if "data" not in t:        
        if "message" in t:
            if ("Could not find summary file" in t["message"]) or ("No data directory could b" in t["message"]):
                return []
        
        raise IOError("night file download fails. Check you authentification maybe?")
    
    return np.sort(t["data"])


    


#######################
#                     #
#  INTERNAL JSON DB   #
#                     #
#######################
# 20181012 20181105
EMPTY_WHAT_DF = pandas.DataFrame(columns=["filename","airmass", "shutter",
                                          "exptime", "target", "night"])

def _parse_line_(line):
    """ """
    try:
        filename, rest = line.split('(')
        info, what = rest.split(")")
        what = what.replace(":", "")
        return [filename.replace(" ","")]+info.split("/")+[what.replace(" [A]","").strip()]
    except:
        return None

def whatfiles_to_dataframe(whatfile):
    """ """
    parsed_lines = [_parse_line_(l_) for l_ in whatfile]
    
    return pandas.DataFrame([l for l in parsed_lines if l is not None],
                                columns=["filename","airmass", "shutter", "exptime", "target"])

def build_datelist(start="2018-06-01", end=None):
    """ """
    if end is None or end in ["today", "now"]:
        from datetime import datetime 
        today = datetime.today()
        end   = today.isoformat().split("T")[0]
            
    return ["%4d%02d%02d"%(tt.year,tt.month, tt.day)
                    for tt in pandas.date_range(start=start, end=end)]


# =============== #
#                 #
#  Pharos IO      #
#                 #
# =============== #

class PharosIO( object ):

    def __init__(self, whatfiles_df=None):
        """ """
        if whatfiles_df is not None:
            self.set_whatdata(whatfiles_df)

    @classmethod
    def load_local(cls, contains="*", stored=False):
        """ load instance from local whatfiles. 

        Parameters
        ----------
        contains: [string] -optional-
            get all the whatfiles containing this. 

        stored: [bool] -optional-
            load the stored multiindex DataFrame. containing
            all the whatfile loaded and then stored.
            If no whatdata has been stored, this will be ignored.

        Returns
        -------
        PharosIO instance
        """
        if stored and os.path.isfile( get_whatfile_path("stored") ):
            return cls( pandas.read_parquet(get_whatfile_path("stored")) )
        
        whatdata =  cls.get_local_whatdata(contains=contains)
        return cls(whatdata)

    #
    # - get_local
    #
    @classmethod
    def get_local_whatdata(cls, contains="*", clean_exptime=True):
        """ get the multiindex dataframe concatenating all the whatfiles.

        Parameters
        ----------
        contains: [string] -optional-
            get all the whatfiles containing this. 
        
        clean_exptime: [bool] -optional-
            by default in the pharos whatfiles exptime entries 
            have the format '{values} s' with the ' s'. 
            If clean_exptime is True, the ' s' is removed

        Returns
        -------
        DataFrame (MultiIndex)
        """
        whatfiles = cls.get_local_whatfiles(contains)
        if len(whatfiles) == 0:
            return None
        
        whatdata= pandas.concat({os.path.basename(f_).split("_")[-1].split(".")[0]:\
                                       pandas.read_parquet(f_)
                         for f_ in whatfiles})
            
        if clean_exptime:
            whatdata["exptime"] = whatdata["exptime"].str.replace(" s", "")

        return whatdata
    
    @staticmethod
    def get_local_whatfiles(contains="*"):
        """ get the file of whatfiles from you computer in $ZTF/sedm/whatfiles.
        = this uses glob =
    
        Parameters
        ----------
        contains: [string] -optional-
            get all the whatfiles containing this. 

        Returns
        -------
        list (path)
        """
        from glob import glob
        return glob(get_whatfile_path(contains))
        
    @staticmethod
    def get_local_pharosfiles(contains="*"):
        """ get the file of pharosfile from you computer in $ZTF/sedm/pharosfiles.

        = this uses glob =
    
        Parameters
        ----------
        contains: [string] -optional-
            get all the whatfiles containing this. 

        Returns
        -------
        list (path)
        """
        from glob import glob
        return glob(get_pharosfile_path(contains))

    @staticmethod
    def _get_missingfile_dates_(which, dates):
        """ get the dates if `which` of the input dates is not in you computer.

        Parameters
        ----------
        which: [string]
            pharosfile or whatfile

        dates: [list of string]
            dates as YYYYMMDD

        Returns
        -------
        list (dates)
        """
        func = eval(f"get_{which}_path")
        return [d_ for d_ in build_datelist() if not os.path.isfile(func(d_))]

    #
    # - download
    #
    @staticmethod
    def download_whatfiles(start="2018-06-01", end=None, client=None, force_dl=False, warn=True):
        """ download the whatfiles in the given time range.

        = set client to use dask multiprocessing =
    
        Paramaters
        ----------
        start: [string] -optional-
            Stating date.

        end: [string or None] -optional-
            End date. 
            If None, it will be today.

        client: [dask Client]  -optional-
            Provide a dask client for using Dask multiprocessing.
            If so, a list of futures will be returned.

        force_dl: [bool] -optional-
            If the file to download already exists locally, shall this re-dlownload it ?

        warn: [bool] -optional-
            If no download are necessary, this will prompt a warning.warn,
            except if warn=False.

        Returns
        -------
        None (or list of futures if client is given)
        """
        dates = build_datelist(start=start, end=end)
        if not force_dl:
            dates = PharosIO._get_missingfile_dates_("whatfile", dates)

        if len(dates)>0:
            return bulk_download("whatfile", np.atleast_1d(dates), client=client)
        if warn:
            warnings.warn("'whatfiles' already up to date")
            
    @staticmethod
    def download_pharosfiles(start="2018-06-01", end=None, client=None, force_dl=False, warn=True):
        """ download the whatfiles in the given time range.

        = set client to use dask multiprocessing =
    
        Paramaters
        ----------
        start: [string] -optional-
            Stating date.

        end: [string or None] -optional-
            End date. 
            If None, it will be today.

        client: [dask Client]  -optional-
            Provide a dask client for using Dask multiprocessing.
            If so, a list of futures will be returned.

        force_dl: [bool] -optional-
            If the file to download already exists locally, shall this re-dlownload it ?

        warn: [bool] -optional-
            If no download are necessary, this will prompt a warning.warn,
            except if warn=False.

        Returns
        -------
        None (or list of futures if client is given)
        """
        dates = build_datelist(start=start, end=end)
        if not force_dl:
            dates = PharosIO._get_missingfile_dates_("pharosfile", dates)

        if len(dates)>0:
            return bulk_download("pharosfile", np.atleast_1d(dates), client=client)
        if warn:
            warnings.warn("'pharosfile' already up to date")


    @staticmethod
    def fetch_pharosfile(date, store=True, load=False, force_dl=False):
        """ """
        if load:
            store=True

        filepath = get_pharosfile_path(date)
        if force_dl or not os.path.isfile(filepath):
            _ = download_pharosfile(date, store=store)
            filepath = get_pharosfile_path(date)
            
        if not load:
            return filepath
        
        return open(filepath).read().splitlines()

    @staticmethod
    def fetch_whatfile(date, store=True, load=False, force_dl=False):
        """ """
        if load:
            store=True
            
        filepath = get_whatfile_path(date)
        if force_dl or not os.path.isfile(filepath):
            _ = download_whatfile(date, store=store, as_dataframe=False)
            filepath = get_whatfile_path(date)
            
        if not load:
            return filepath
        
        return pandas.read_parquet(filepath)
    
    #
    # - Storing
    #
    def store(self):
        """ Store locally $ZTF/sedm/whatfiles the current whatdata. """
        self.whatdata.to_parquet( get_whatfile_path("stored") )
        
    def update(self, force_dl=False, store=True, client=None):
        """ update the instance:
        1. call download_whatfiles() to get the missing whatfiles if any
        2. Reload all the whatfiles to rebuild whatdata.
        3. Store the updated whatdata (if store=True)

        Parameters
        ----------

        // download_whatfiles options

        force_dl: [bool] -optional-
            If the file to download already exists locally, shall this re-dlownload it ?

        client: [dask Client]  -optional-
            Provide a dask client for using Dask multiprocessing.
            If so, a list of futures will be returned.
            
        // other
        store: [bool] -optional-
            Shall the updated whatdata be stored (using store())
        
        
        """
        f_if_any = self.download_whatfiles(force_dl=force_dl, client=client)
        if client is not None: # wait for the end before reloading.
            from dask.distributed import wait
            _ = wait(f_if_any)
            
        self.set_whatdata(self.get_local_whatdata())
        if store:
            self.store()
        
    # ============== #
    #  Methods       #
    # ============== #
    # -------- #
    #  SETTER  #
    # -------- #    
    # -------- #
    #  SETTER  #
    # -------- #
    def set_whatdata(self, what_dataframe, infer_type=True):
        """ """
        if infer_type:
            what_dataframe = what_dataframe.astype({"airmass":"float",
                                                    "shutter":"float",
                                                    "exptime":"float"})
        self._whatdata = what_dataframe

    # -------- #
    #  GETTER  #
    # -------- #
    def get_whatdata(self, targets=None,
                         incl_calib=False, incl_std=True, incl_ztf=True,
                         calib_only=False, std_only=False, ztf_only=False,
                         airmass_range=None, exptime_range=None, date_range=None):
        """ Get the subpart of the whatdata DataFrame
        
        Parameters
        ----------
        targets: [string or list of] -optional-
            Select only whatdata entries corresponding to this target (or these targets).
            
        incl_calib, incl_std, incl_ztf: [bool] -optional-
            Include the calibration, standard star observations and ztf observations ?

        calib_only, std_only, ztf_only: [bool] -optional-
            Limit the returned entries as the one corresponding to 
            calibration, standard stars or ztf names targets.

        airmass_range, exptime_range, date_range: [float/None, float/None] -optional-
            select a range or airmass and/or exptime.
            None means no limit. For instance all the exposure lower than 300s
            will be exptime_range=[None,300]
            * Format * for date_range, dates are given as YYYYMMDD
            
        Returns
        -------
        DataFrame (subpart of whatdata)
        """
        def _parse_range_(din, key, krange):
            # Internal
            kmin, kmax = krange
            if kmin is None:
                return din[din[key].astype("float")<kmax]
            if kmax is None:
                return din[din[key].astype("float")>kmin]
            return din[din[key].astype("float").between(kmin, kmax)]

        data_ = self.whatdata.copy()
        if targets is not None:
            data_ = data_[data_["target"].isin(np.atleast_1d(targets))]
            
        if not incl_calib:
            data_ = data_[~data_["target"].str.startswith("Calib")]
            
        if not incl_std:
            data_ = data_[~data_["target"].str.startswith("STD")]
            
        if not incl_ztf:
            data_ = data_[~data_["target"].str.startswith("ZTF")]

        if std_only:
            data_ = data_[data_["target"].str.startswith("STD")]

        if ztf_only:
            data_ = data_[data_["target"].str.startswith("ZTF")]
            
        if calib_only:
            data_ = data_[data_["target"].str.startswith("Calib")]
            
        if exptime_range is not None:
            data_ = _parse_range_(data_, "exptime", exptime_range)

        if airmass_range is not None:
            data_ = _parse_range_(data_, "airmass", airmass_range)

        if date_range is not None:
            kmin, kmax = date_range
            dates = np.asarray(data_.index.get_level_values(0), dtype="float")
            
            if kmin is None:
                flag = dates<float(kmax)
            elif kmax is None:
                flag = dates>float(kmin)
            else:
                flag = (dates<float(kmax)) * (dates>float(kmin))
            data_ = data_[flag]

        return data_

    def get_pharosdata(self, date, force_dl=False, basename=False,
                           add_guider=True, add_snid=True):
        """ """
        if date not in self.pharosdata or force_dl:
            d = self.fetch_pharosfile(date, force_dl=force_dl, load=True)
            d += [f.replace("bkgd_","") for f in d if os.path.basename(f).startswith("bkgd_crr")]
            self.pharosdata[date] = d

        if basename:
            pdata = [os.path.basename(l) for l in self.pharosdata[date]]
        else:
            pdata = self.pharosdata[date].copy()


        if add_guider:
            targetcubes = [l for l in pdata if os.path.basename(l).startswith("e3d")
                               and "_STD-" not in l]
            guiders = [self.build_cube_astromfile(cubef) for cubef in targetcubes]
            pdata += guiders

        if add_snid:
            targetspecs = [l for l in pdata if os.path.basename(l).startswith("spec")
                               and "_STD-" not in l and l.endswith(".fits")]
            snidout = [self._specname_to_snidpath_(s_) for s_ in targetspecs]
            pdata += snidout

        return pdata
        

    def get_nights_with_target(self, targetname, 
                                  airmass_range=None,
                                  exptime_range=None,
                                  date_range=None, **kwargs):
        """ """
        target_what = self.get_whatdata(targets=targetname,
                                        airmass_range=airmass_range, 
                                        exptime_range=exptime_range,
                                        date_range=date_range,
                                            **kwargs)
        return np.unique(target_what.index.get_level_values(0))
    
    def get_target_pharosdata(self, targetname,
                                  airmass_range=None,
                                  exptime_range=None,
                                  date_range=None,
                                  kind=None,
                                  contains=None, not_contains=None,
                                  extension=None,
                                  add_guider=True, add_snid=True,
                                  **kwargs):
        """ """
        targetbasename = self.get_target_basename(targetname=targetname,
                                        airmass_range=airmass_range, 
                                        exptime_range=exptime_range,
                                        date_range=date_range,
                                            **kwargs)
        
        all_dates = np.unique(targetbasename.index.get_level_values(0))
        
        files = {}
        for date_ in all_dates:
            tfile = np.asarray(targetbasename.xs(date_)["basename"].values, dtype="str")
            pdata = self.get_pharosdata(date_, force_dl=False,
                                        add_guider=add_guider, add_snid=add_snid)
            files_ = [os.path.basename(l) for l in pdata if np.any([k in l for k in tfile])]
            # not fastest, but it's very fast anyway ; it's easier to read this way
            if not (extension is None or extension in ["*","any", "all"]):
                files_ = [f_ for f_ in files_ if f_.endswith(extension)]
                
            if not (kind is None or kind in ["*","any","all"]):
                files_ = [f_ for f_ in files_ if f_.startswith(kind)]
                
            if not (contains is None or contains in ["*","any","all"]):
                files_ = [f_ for f_ in files_ if contains in f_]

            if not_contains is not None:
                files_ = [f_ for f_ in files_ if not not_contains in f_]
                
            files[date_] = files_

        
        return files
        
    def get_target_basename(self, targetname, date_range=None, **kwargs):
        """ get a dataframe the basenames assocated to the target """
        target_what = self.get_whatdata(targets=targetname,
                                        date_range=date_range,
                                        **kwargs)
        return pandas.DataFrame(target_what["filename"].astype("str"
                        ).str.split(".", expand=True).rename({0:"basename"}, axis=1)["basename"])


    # -------------- #
    #  Internal      #
    # -------------- #
    @classmethod
    def build_cube_astromfile(cls, cubename):
        """ """
        basename = cls._cubename_to_basename_(cubename)
        return cls.build_guiderpath(basename)
    
    @staticmethod
    def _cubename_to_basename_(cubename):
        """ """
        return "_".join(os.path.basename(cubename).split("_b_")[-1].split("_")[:4])
    
    @staticmethod
    def _specname_to_snidpath_(specname):
        """ """
        ext = specname.split(".")[-1]
        return specname.replace(".fits", "_snid.output")

    @classmethod
    def build_guiderpath(cls, basename, astrom=True):
        """ """
        date = cls.basename_to_date(basename)
        ext = "_astrom" if astrom else ""
        return os.path.join("/data", date, f"guider_crr_b_{basename}{ext}.fits.gz")

    @staticmethod
    def basename_to_date(basename):
        """ """
        return basename.split("_")[0].replace("ifu","")

    # ============== #
    #  Properties    #
    # ============== #
    @property
    def whatdata(self):
        """ (MultiIndex) DataFrame containing the whatfiles. """
        if not hasattr(self,"_whatdata"):
            return None
        return self._whatdata
    
    def has_whatdata(self):
        """ """
        return self.whatdata is not None

    @property
    def pharosdata(self):
        """ dict containing the file available for a given date."""
        if not hasattr(self, "_pharosdata") or self._pharosdata is None:
            self._pharosdata = {}
        return self._pharosdata

    def has_pharosdata():
        """ """
        return len(self.pharosdata)>0
    
            

                
        
  
##################
#                #
#  PHAROS        #
#                #
##################
class SEDMQuery():
    def __init__(self, load_pharosio=True, **kwargs):
        """ """
        if load_pharosio:
            self._load_pharosio_(**kwargs)
    
    def _load_pharosio_(self, **kwargs):
        """ """
        self._pharosio = PharosIO.load_local(stored=True, **kwargs)

    # ============= #
    #  Method       #
    # ============= #
    def update_pharosio(self, client=None, **kwargs):
        """ """
        return self.pharosio.update(client=client, **kwargs)
    
    def get_whatdata(self, targets=None,
                     incl_calib=False, incl_std=True, incl_ztf=True,
                     calib_only=False, std_only=False, ztf_only=False,
                     airmass_range=None, exptime_range=None, date_range=None):
        """ Get the subpart of the whatdata DataFrame
        
        Parameters
        ----------
        targets: [string or list of] -optional-
            Select only whatdata entries corresponding to this target (or these targets).
            
        incl_calib, incl_std, incl_ztf: [bool] -optional-
            Include the calibration, standard star observations and ztf observations ?

        calib_only, std_only, ztf_only: [bool] -optional-
            Limit the returned entries as the one corresponding to 
            calibration, standard stars or ztf names targets.

        airmass_range, exptime_range, date_range: [float/None, float/None] -optional-
            select a range or airmass and/or exptime.
            None means no limit. For instance all the exposure lower than 300s
            will be exptime_range=[None,300]
            * Format * for date_range, dates are given as YYYYMMDD
            
        Returns
        -------
        DataFrame (subpart of whatdata)
        """
        klocal = locals().copy()
        klocal.pop("self")
        return self.pharosio.get_whatdata(**klocal)
    
    def get_targetnames(self, **kwargs):
        """ """
        whatdata = self.get_whatdata(**kwargs)
        return np.asarray(np.unique(whatdata["target"]), dtype="str")

    # ============= #
    #  Fetch        #
    # ============= #
    # ------- #
    #  Night  #
    # ------- #
    def get_night_fluxcal(self, date,
                              download_missing=True, exist=True, 
                              contains=None, not_contains=None, client=None, **kwargs):
        """ date could be a list """
        return self.get_night_data(date, kind="fluxcal", extension=".fits",
                                download_missing=download_missing, exist=exist, 
                                contains=contains, not_contains=not_contains, client=client,
                                **kwargs)

    def get_night_crr(self, date, contains=None, not_contains=None,
                            client=None, force_dl=False,
                            ioprop={}, **kwargs):
        """ """
        prop = dict(kind="crr", contains=contains, not_contains=not_contains, extension=".fits")
        return self.get_night_data(date, client=client, ioprop=ioprop,
                                    force_dl=force_dl, **{**prop,**kwargs})

    def get_night_cubes(self, date, contains=None, not_contains="e3d_dome",
                            client=None, force_dl=False,
                            incl_dome=False,
                            ioprop={}, **kwargs):
        """ """
        prop = dict(kind="e3d", contains=contains, not_contains=not_contains, extension=".fits")   
        return self.get_night_data(date, client=client, ioprop=ioprop,
                                    force_dl=force_dl, **{**prop,**kwargs})

    def get_night_spectra(self, date, contains=None, not_contains=None, client=None, force_dl=False,
                            incl_dome=False, ioprop={}, **kwargs):
        """ """
        prop = dict(kind="spec", contains=contains, not_contains=not_contains, extension=".fits")
        return self.get_night_data(date, client=client, ioprop=ioprop, force_dl=force_dl,
                                       **{**prop,**kwargs})

    def get_night_snidout(self, date, contains=None, not_contains=None, client=None, force_dl=False,
                            incl_dome=False, ioprop={}, **kwargs):
        """ """
        prop = dict(kind="spec", contains=contains, not_contains=not_contains, extension="snid.output")
        return self.get_night_data(date, client=client, ioprop=ioprop, force_dl=force_dl,
                                       **{**prop,**kwargs})

    def get_night_astrom(self, date, not_contains=None, client=None, force_dl=False,
                            incl_dome=False, ioprop={}, **kwargs):
        """ """
        prop = dict(kind="guider", contains="astrom", not_contains=not_contains, extension="*")
        return self.get_night_data(date, client=client, ioprop=ioprop, force_dl=force_dl,
                                       **{**prop,**kwargs})

    def get_night_data(self, date, download_missing=True, exist=True, force_dl=False,
                           contains=None, not_contains=None, kind=None, extension="*", 
                           client=None, ioprop={}, **kwargs):
        """ """
        pharosdata = {date_:
                      [l for l in self.pharosio.get_pharosdata(date_, basename=True, **ioprop) 
                        if  (kind is None or l.startswith(kind)) and 
                            (extension is None or extension in ["*","all"] or \
                                 l.endswith(extension)) and 
                            (contains is None or contains in l) and 
                            (not_contains is None or not_contains not in l)]
                                
                     for date_ in np.atleast_1d(date)}


        local_path = np.concatenate(self._pharosdata_to_datapath_(pharosdata, "local"))
        
        if download_missing or force_dl:
            index_missing = [not os.path.isfile(l) for l in local_path]
            if np.any(index_missing) or force_dl:
                if force_dl:
                    index_missing=None
                futures_ifany = self._download_pharosdata_(pharosdata, index=index_missing,
                                                               client=client, **kwargs)
                if client is not None:
                    from dask.distributed import wait
                    _ = wait(futures_ifany)
                    
        if exist:
            return [l for l in local_path if os.path.isfile(l)]
        return local_path
    
    # ------- #
    # Target  #
    # ------- #
    def get_target_snidoutput(self, targetname, download_missing=True, exist=True, force_dl=False,
                          contains=None, not_contains=None, client=None, ioprop={}, **kwargs):
        """ 
        Parameters:
        -----------
        ioprop: [dict] -optional-
            used as kwargs for get_whatdata().
            Incl: incl_calib, incl_std, incl_ztf, 
                  calib_only, std_only, ztf_only,
                  airmass_range, exptime_range, date_range

        """
        prop = dict(kind="spec", contains=contains, not_contains=not_contains,
                    extension="snid.output")
        return self.get_target_datapath(targetname, client=client, ioprop=ioprop,
                                            force_dl=force_dl, **{**prop,**kwargs})
    
    
    def get_target_fluxcal(self, targetname, download_missing=True, exist=True,
                                     contains=None, not_contains=None, client=None,
                                    **kwargs):
        """ get nights associated to the target and downloads the fluxcal files of that night.
        uses self.get_night_fluxcal()
        """
        dates = self.pharosio.get_nights_with_target(targetname, **kwargs)
        return self.get_night_fluxcal(dates,
                                    download_missing=download_missing, exist=exist, 
                                    contains=contains, not_contains=not_contains,
                                    client=client, **kwargs)

    def get_target_astrom(self, targetname, download_missing=True, exist=True, force_dl=False,
                          not_contains=None, client=None, ioprop={}, **kwargs):
        """ 
        Parameters:
        -----------
        ioprop: [dict] -optional-
            used as kwargs for get_whatdata().
            Incl: incl_calib, incl_std, incl_ztf, 
                  calib_only, std_only, ztf_only,
                  airmass_range, exptime_range, date_range

        """
        prop = dict(kind="guider", contains="astrom", not_contains=not_contains, extension="*")
        return self.get_target_datapath(targetname, client=client, ioprop=ioprop,
                                            force_dl=force_dl, **{**prop,**kwargs})

    def get_target_guider(self, targetname, download_missing=True, exist=True, force_dl=False,
                          contains=None, not_contains=None, client=None, ioprop={}, **kwargs):
        """ 
        Parameters:
        -----------
        ioprop: [dict] -optional-
            used as kwargs for get_whatdata().
            Incl: incl_calib, incl_std, incl_ztf, 
                  calib_only, std_only, ztf_only,
                  airmass_range, exptime_range, date_range

        """
        prop = dict(kind="guider", contains=contains, not_contains=not_contains, extension=".fits")
        return self.get_target_datapath(targetname, client=client,ioprop=ioprop,
                                            force_dl=force_dl, **{**prop,**kwargs})

    def get_target_spectra(self, targetname, download_missing=True, exist=True, force_dl=False,
                          contains=None, not_contains=None, client=None, ioprop={}, **kwargs):
        """ 
        Parameters:
        -----------
        ioprop: [dict] -optional-
            used as kwargs for get_whatdata().
            Incl: incl_calib, incl_std, incl_ztf, 
                  calib_only, std_only, ztf_only,
                  airmass_range, exptime_range, date_range

        """
        prop = dict(kind="spec", contains=contains, not_contains=not_contains, extension=".fits")
        return self.get_target_datapath(targetname, client=client,ioprop=ioprop,
                                            force_dl=force_dl, **{**prop,**kwargs})
    
    def get_target_cubes(self, targetname, download_missing=True, exist=True, force_dl=False,
                          contains=None, not_contains=None, client=None,ioprop={},  **kwargs):
        """ 
        Parameters:
        -----------
        ioprop: [dict] -optional-
            used as kwargs for get_whatdata().
            Incl: incl_calib, incl_std, incl_ztf, 
                  calib_only, std_only, ztf_only,
                  airmass_range, exptime_range, date_range
        """
        prop = dict(kind="e3d", contains=contains, not_contains=not_contains, extension=".fits")
        return self.get_target_datapath(targetname, client=client, ioprop=ioprop,
                                            force_dl=force_dl, **{**prop,**kwargs})

    def get_target_crr(self, targetname, download_missing=True, exist=True, force_dl=False,
                          contains=None, not_contains=None, client=None, ioprop={},  **kwargs):
        """ 
        Parameters:
        -----------
        ioprop: [dict] -optional-
            used as kwargs for get_whatdata().
            Incl: incl_calib, incl_std, incl_ztf, 
                  calib_only, std_only, ztf_only,
                  airmass_range, exptime_range, date_range
        """
        prop = dict(kind="crr", contains=contains, not_contains=not_contains, extension=".fits")
        return self.get_target_datapath(targetname, client=client, ioprop=ioprop,
                                            force_dl=force_dl, **{**prop,**kwargs})

    
    def get_target_datapath(self, targetname, kind, contains=None, not_contains=None, 
                            download_missing=True, exist=True, force_dl=False, verbose=False,
                            extension=".fits",  ioprop={}, client=None, **kwargs):
        """ 

        Parameters:
        -----------
        ioprop: [dict] -optional-
            used as kwargs for get_whatdata().
            Incl: incl_calib, incl_std, incl_ztf, 
                  calib_only, std_only, ztf_only,
                  airmass_range, exptime_range, date_range

        """
        prop = dict(contains=contains, not_contains=not_contains, extension=extension)
        if verbose:
            print(prop)
        pharosdata = self.pharosio.get_target_pharosdata(targetname, kind=kind,
                                                             **{**prop, **ioprop})
        
        if pharosdata is None or len(pharosdata)==0:
            return None
        
        local_path = np.concatenate(self._pharosdata_to_datapath_(pharosdata, "local"))
        if verbose:
            print(local_path)
            
        if download_missing or force_dl:
            index_missing = [not os.path.isfile(l) for l in local_path]
            if verbose:
                print(f"missing, {index_missing}")
            if np.any(index_missing) or force_dl:
                if force_dl:
                    index_missing=None
                    
                futures_ifany = self._download_pharosdata_(pharosdata, index=index_missing,
                                                               client=client, **kwargs)
                if client is not None:
                    from dask.distributed import wait
                    _ = wait(futures_ifany)
                    
        if exist:
            return [l for l in local_path if os.path.isfile(l)]
        return local_path
        
    # ============= #
    #  Download     #
    # ============= #
    # -------- #
    #  Night   #
    # -------- #
    def bulk_download_night(self, date, contains=None, not_contains=None, kind=None, extension="*", 
                                dirout=None, client=None, nprocess=None, **kwargs):
        """ """
        pharosdata = {date_:
                      [l for l in self.pharosio.get_pharosdata(date_, basename=True) 
                        if  (kind is None or l.startswith(kind)) and 
                            (extension is None or extension in ["*","all"] or l.endswith(extension)) and 
                            (contains is None or contains in l) and 
                            (not_contains is None or not_contains not in l)]
                                
                     for date_ in np.atleast_1d(date)}
        
        return self._download_pharosdata_(pharosdata, dirout=dirout, nprocess=nprocess, client=client, 
                                            **kwargs)
        
    def download_night_calibrations(self, date, which="*", 
                                      dirout=None, client=None, nprocess=4, **kwargs):
        """ 
        Parameters
        ----------
        date: [string (or list of)]
            night date as YYYYMMDD. It could be a list or [YYYYMMDD, YYYYMMDD etc.]
            
        which: [string or list of]
            kind on calibration file to download. 
            could any (of list of): 
            - ['HexaGrid.pkl', 'TraceMatch.pkl','TraceMatch_WithMasks.pkl', 
               'WaveSolution.pkl','Flatfits']
            - which='*' means all of them.
            
        // download options
        
        dirout: [string] -optional-
            where the data should be downloaded (will actually be dirout/YYYYMMDD/_downloaded_data_).
            by default this will be `$ZTDATA/sedm/redux`.
                        
        client: [Dask Client] -optional-
            Dask client used for the downloaded. 
            If so, a list of futures will be returned.
            
        nprocess: [int] -optional-
            number of parallel downloading. 
            - ignored if client is given -
        
        **kwargs goes to download_target_data()
         incl: ignore_warnings=False, verbose=False, overwrite=True
            nodl: [bool] -optional-
                if nodl=True, nothing is downloaded and
                the list_of_url and list_of_path_where_downloaded are returned.
            
            show_progress: [bool] -optional-
                show the progress of downloading
                - ignored if client is used -
        """
        NIGHTLY = ["HexaGrid.pkl", "TraceMatch.pkl","TraceMatch_WithMasks.pkl", "WaveSolution.pkl","Flat.fits"]
        
        if which in ["*","all"]:
            which = NIGHTLY
        elif np.any([w_ not in NIGHTLY for w_ in np.atleast_1d(which)]):
            raise ValueError(f"At least one of the which input is not a known nighly calibration: {which}")
            
        pharosdata = {}
        for date_ in np.atleast_1d(date):
            contains = [f"{date_}_{k}" for k in np.atleast_1d(which)]
            pharosdata[date_] = [l for l in self.pharosio.get_pharosdata(date_, basename=True)
                                    if np.any([k in l for k in contains])]
        
        return self._download_pharosdata_(pharosdata, dirout=dirout, nprocess=nprocess, client=client, 
                                          **kwargs)
    
    def download_night_fluxcal(self, date, contains=None, not_contains=None,
                                      dirout=None, client=None, nprocess=4, **kwargs):
        """ 
        
        Parameters
        ----------
        date: [string (or list of)]
            night date as YYYYMMDD. It could be a list or [YYYYMMDD, YYYYMMDD etc.]
            
        contains: [string] -optional-
            element of the name that should be in the pharos' filename.
            
        not_contains: [string] -optional-
            element of the name that should *not* be in the pharos' filename.
    
        // download options
        
        dirout: [string] -optional-
            where the data should be downloaded (will actually be dirout/YYYYMMDD/_downloaded_data_).
            by default this will be `$ZTDATA/sedm/redux`.
                        
        client: [Dask Client] -optional-
            Dask client used for the downloaded. 
            If so, a list of futures will be returned.
            
        nprocess: [int] -optional-
            number of parallel downloading. 
            - ignored if client is given -
        
        **kwargs goes to download_target_data()
         incl: ignore_warnings=False, verbose=False, overwrite=True
            nodl: [bool] -optional-
                if nodl=True, nothing is downloaded and
                the list_of_url and list_of_path_where_downloaded are returned.
            
            show_progress: [bool] -optional-
                show the progress of downloading
                - ignored if client is used -

        Returns
        -------
        
        """
        pharosdata = {date_:[l for l in self.pharosio.get_pharosdata(date_, basename=True)
                            if l.startswith("fluxcal") and 
                            (contains is None or contains in l) and 
                            (not_contains is None or not_contains not in l)]
                     for date_ in np.atleast_1d(date)}
        
        return self._download_pharosdata_(pharosdata, dirout=dirout, nprocess=nprocess, client=client, 
                                          **kwargs)
            
    # -------- #
    #  Target  #
    # -------- #
    def download_target_astrom(self, targetname, not_contains=None, 
                               dirout=None, client=None, nprocess=4, ioprop={},**kwargs):
        """ download cubes associated to a target. 
        = uses download_target_data() with kind='e3d' =
        
        Parameters
        ----------
        targetname: [string]
            name to the target for which you want to download data
            
        contains: [string] -optional-
            element of the name that should be in the pharos' filename.
            
        not_contains: [string] -optional-
            element of the name that should *not* be in the pharos' filename.
            
        // donwload options
        
        dirout: [string] -optional-
            where the data should be downloaded (will actually be dirout/YYYYMMDD/_downloaded_data_).
            by default this will be `$ZTDATA/sedm/redux`.
            
        client: [Dask Client] -optional-
            Dask client used for the downloaded. 
            If so, a list of futures will be returned.
            
        nprocess: [int] -optional-
            number of parallel downloading. 
            - ignored if client is given -
        
        // whatdata options

        ioprop: [dict] -optional-
            used as kwargs for get_whatdata().
            Incl: incl_calib, incl_std, incl_ztf, 
                  calib_only, std_only, ztf_only,
                  airmass_range, exptime_range, date_range



        **kwargs goes to download_target_data()
         incl: nodl=False, ignore_warnings=False, show_progress=True, verbose=False, overwrite=True
            nodl: [bool] -optional-
                if nodl=True, nothing is downloaded and
                the list_of_url and list_of_path_where_downloaded are returned.
            
            show_progress: [bool] -optional-
                show the progress of downloading
                - ignored if client is used -s
        Returns
        -------
        None (or list of futures if client given)
        """
        prop = dict(kind="guider", contains="astrom", not_contains=not_contains, extension=".fits")
        return self.download_target_data(targetname, dirout=dirout, client=client, nprocess=nprocess,
                                             ioprop=ioprop,
                                         **{**prop,**kwargs})
    
    def download_target_guider(self, targetname, contains=None, not_contains=None, 
                               dirout=None, client=None, nprocess=4, ioprop={}, **kwargs):
        """ download cubes associated to a target. 
        = uses download_target_data() with kind='e3d' =
        
        Parameters
        ----------
        targetname: [string]
            name to the target for which you want to download data
            
        contains: [string] -optional-
            element of the name that should be in the pharos' filename.
            
        not_contains: [string] -optional-
            element of the name that should *not* be in the pharos' filename.
            
        // download options

        dirout: [string] -optional-
            where the data should be downloaded (will actually be dirout/YYYYMMDD/_downloaded_data_).
            by default this will be `$ZTDATA/sedm/redux`.
            
        client: [Dask Client] -optional-
            Dask client used for the downloaded. 
            If so, a list of futures will be returned.
            
        nprocess: [int] -optional-
            number of parallel downloading. 
            - ignored if client is given -
        
        // whatdata options

        ioprop: [dict] -optional-
            used as kwargs for get_whatdata().
            Incl: incl_calib, incl_std, incl_ztf, 
                  calib_only, std_only, ztf_only,
                  airmass_range, exptime_range, date_range


        **kwargs goes to download_target_data()
         incl: nodl=False, ignore_warnings=False, show_progress=True, verbose=False, overwrite=True
            nodl: [bool] -optional-
                if nodl=True, nothing is downloaded and
                the list_of_url and list_of_path_where_downloaded are returned.
            
            show_progress: [bool] -optional-
                show the progress of downloading
                - ignored if client is used -
        Returns
        -------
        None (or list of futures if client given)
        """
        prop = dict(kind="guider", contains=contains, not_contains=not_contains, extension=".fits")
        return self.download_target_data(targetname,dirout=dirout, client=client, nprocess=nprocess,
                                             ioprop=ioprop,
                                         **{**prop,**kwargs})
    
    def download_target_spectra(self, targetname, contains=None, not_contains=None, 
                                dirout=None, client=None, nprocess=4,ioprop={}, **kwargs):
        """ download spectra associated to a target. 
        = uses download_target_data() with kind='e3d' =
        
        Parameters
        ----------
        targetname: [string]
            name to the target for which you want to download data
            
        contains: [string] -optional-
            element of the name that should be in the pharos' filename.
            
        not_contains: [string] -optional-
            element of the name that should *not* be in the pharos' filename.
            
        // download options

        dirout: [string] -optional-
            where the data should be downloaded (will actually be dirout/YYYYMMDD/_downloaded_data_).
            by default this will be `$ZTDATA/sedm/redux`.
            
        client: [Dask Client] -optional-
            Dask client used for the downloaded. 
            If so, a list of futures will be returned.
            
        nprocess: [int] -optional-
            number of parallel downloading. 
            - ignored if client is given -
        
        // whatdata options

        ioprop: [dict] -optional-
            used as kwargs for get_whatdata().
            Incl: incl_calib, incl_std, incl_ztf, 
                  calib_only, std_only, ztf_only,
                  airmass_range, exptime_range, date_range


        **kwargs goes to download_target_data()
         incl: nodl=False, ignore_warnings=False, show_progress=True, verbose=False, overwrite=True
            nodl: [bool] -optional-
                if nodl=True, nothing is downloaded and
                the list_of_url and list_of_path_where_downloaded are returned.
            
            show_progress: [bool] -optional-
                show the progress of downloading
                - ignored if client is used -
                
        Returns
        -------
        None (or list of futures if client given)
        """
        prop = dict(kind="spec", contains=contains, not_contains=not_contains, extension=".fits")
        return self.download_target_data(targetname, dirout=dirout, client=client, nprocess=nprocess,
                                             ioprop=ioprop,                                             
                                         **{**prop,**kwargs})
    
    def download_target_cubes(self, targetname, contains=None, not_contains=None, 
                              dirout=None, client=None, nprocess=4, ioprop={}, **kwargs):
        """ download cubes associated to a target. 
        = uses download_target_data() with kind='e3d' =
        
        Parameters
        ----------
        targetname: [string]
            name to the target for which you want to download data
            
        contains: [string] -optional-
            element of the name that should be in the pharos' filename.
            
        not_contains: [string] -optional-
            element of the name that should *not* be in the pharos' filename.
            
        // download options

        dirout: [string] -optional-
            where the data should be downloaded (will actually be dirout/YYYYMMDD/_downloaded_data_).
            by default this will be `$ZTDATA/sedm/redux`.
            
        client: [Dask Client] -optional-
            Dask client used for the downloaded. 
            If so, a list of futures will be returned.
            
        nprocess: [int] -optional-
            number of parallel downloading. 
            - ignored if client is given -
        
        // whatdata options

        ioprop: [dict] -optional-
            used as kwargs for get_whatdata().
            Incl: incl_calib, incl_std, incl_ztf, 
                  calib_only, std_only, ztf_only,
                  airmass_range, exptime_range, date_range

        **kwargs goes to download_target_data()
         incl: nodl=False, ignore_warnings=False, show_progress=True, verbose=False, overwrite=True
            nodl: [bool] -optional-
                if nodl=True, nothing is downloaded and
                the list_of_url and list_of_path_where_downloaded are returned.
            
            show_progress: [bool] -optional-
                show the progress of downloading
                - ignored if client is used -
                
        Returns
        -------
        None (or list of futures if client given)
        """
        prop = dict(kind="e3d", contains=contains, not_contains=not_contains, extension=".fits")
        return self.download_target_data(targetname, dirout=dirout, client=client,
                                             nprocess=nprocess,
                                             ioprop=ioprop,
                                         **{**prop,**kwargs})
        
    def download_target_fluxcal(self, targetname, contains=None, not_contains=None, 
                                 dirout=None, client=None, nprocess=4, **kwargs):
        """ """
        dates = self.pharosio.get_nights_with_target(targetname, **kwargs)
        return self.download_night_fluxcal(dates, contains=contains, not_contains=not_contains, 
                                           dirout=dirout, nprocess=nprocess, client=client,
                                           **kwargs)
        
    def download_target_data(self, targetname, kind, 
                             contains=None, not_contains=None, extension=".fits", 
                             dirout=None, nodl=False, 
                             show_progress=True, nprocess=4, client=None, ioprop={}, **kwargs):
        """ 

        // whatdata options

        ioprop: [dict] -optional-
            used as kwargs for get_whatdata().
            Incl: incl_calib, incl_std, incl_ztf, 
                  calib_only, std_only, ztf_only,
                  airmass_range, exptime_range, date_range


        """
        prop = dict(contains=contains, not_contains=not_contains, extension=extension)
        pharosdata = self.pharosio.get_target_pharosdata(targetname, kind=kind,
                                                             **{**prop, **ioprop})
        return self._download_pharosdata_(pharosdata, 
                                          dirout=dirout, nodl=nodl, 
                                          show_progress=show_progress, nprocess=nprocess,
                                          client=client,
                                         **kwargs)
        
    def _download_pharosdata_(self, pharosdata, dirout=None,
                             nodl=False, ignore_warnings=False,
                             show_progress=True, verbose=False, overwrite=True, 
                             nprocess=4, client=None, index=None):
        """ """
        if dirout is None:
            dirout = "local"
        
        pharos_path = np.concatenate(self._pharosdata_to_datapath_(pharosdata, "pharos"))
        local_path = np.concatenate(self._pharosdata_to_datapath_(pharosdata, dirout))
        nprocess = np.min([ len(pharos_path), nprocess] )
        if index is not None:
            pharos_path = pharos_path[index]
            local_path = local_path[index]
            
        if nodl:
            return pharos_path, local_path
        
        with warnings.catch_warnings():
            if ignore_warnings:
                warnings.simplefilter("ignore")
            
            return io.download_url(pharos_path, local_path,
                        show_progress = show_progress, verbose=verbose,
                        overwrite=overwrite, nprocess=nprocess, 
                        cookies="no_cookies",client=client,
                        filecheck=False)
        
    # ============= #
    #  Internal     #
    # ============= #
    @staticmethod
    def _pharosdata_to_datapath_(pharosdata, source):
        """ """
        if source == "local":
            source = os.path.join(SEDMLOCAL_BASESOURCE, "redux")
        elif source == "pharos":
            source = os.path.join(PHAROS_BASEURL, "data")
        elif source is None:
            source = ""
            
        return [[os.path.join(source,k,v_) for v_ in v] for k,v in pharosdata.items()]
        
    # ============= #
    #  Properties   #
    # ============= #
    @property
    def pharosio(self):
        """ """
        return self._pharosio
